# =============================================================================
# Copyright 2024 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from typing import Optional

import lightning.pytorch as pl
from loguru import logger
from torch.utils.data import DataLoader, Dataset
import yaml

from ptlflow.data import flow_transforms as ft
from ptlflow.data.datasets import (
    AutoFlowDataset,
    FlyingChairsDataset,
    FlyingChairs2Dataset,
    Hd1kDataset,
    KittiDataset,
    KubricDataset,
    MiddleburySTDataset,
    SintelDataset,
    FlyingThings3DDataset,
    FlyingThings3DSubsetDataset,
    SpringDataset,
    TartanAirDataset,
    ViperDataset,
)
from ptlflow.utils.utils import make_divisible


class FlowDataModule(pl.LightningDataModule):
    def __init__(
        self,
        predict_dataset: Optional[str] = None,
        test_dataset: Optional[str] = None,
        train_dataset: Optional[str] = None,
        val_dataset: Optional[str] = None,
        train_batch_size: Optional[int] = None,
        train_num_workers: int = 4,
        train_crop_size: tuple[int, int] = None,
        train_transform_cuda: bool = False,
        train_transform_fp16: bool = False,
        autoflow_root_dir: Optional[str] = None,
        flying_chairs_root_dir: Optional[str] = None,
        flying_chairs2_root_dir: Optional[str] = None,
        flying_things3d_root_dir: Optional[str] = None,
        flying_things3d_subset_root_dir: Optional[str] = None,
        mpi_sintel_root_dir: Optional[str] = None,
        kitti_2012_root_dir: Optional[str] = None,
        kitti_2015_root_dir: Optional[str] = None,
        hd1k_root_dir: Optional[str] = None,
        tartanair_root_dir: Optional[str] = None,
        spring_root_dir: Optional[str] = None,
        kubric_root_dir: Optional[str] = None,
        dataset_config_path: str = "./datasets.yaml",
    ):
        super().__init__()
        self.predict_dataset = predict_dataset
        self.test_dataset = test_dataset
        self.train_dataset = train_dataset
        self.val_dataset = val_dataset
        self.train_batch_size = train_batch_size
        self.train_num_workers = train_num_workers
        self.train_crop_size = train_crop_size
        self.train_transform_cuda = train_transform_cuda
        self.train_transform_fp16 = train_transform_fp16

        self.autoflow_root_dir = autoflow_root_dir
        self.flying_chairs_root_dir = flying_chairs_root_dir
        self.flying_chairs2_root_dir = flying_chairs2_root_dir
        self.flying_things3d_root_dir = flying_things3d_root_dir
        self.flying_things3d_subset_root_dir = flying_things3d_subset_root_dir
        self.mpi_sintel_root_dir = mpi_sintel_root_dir
        self.kitti_2012_root_dir = kitti_2012_root_dir
        self.kitti_2015_root_dir = kitti_2015_root_dir
        self.hd1k_root_dir = hd1k_root_dir
        self.tartanair_root_dir = tartanair_root_dir
        self.spring_root_dir = spring_root_dir
        self.kubric_root_dir = kubric_root_dir
        self.dataset_config_path = dataset_config_path

        self.predict_dataset_parsed = None
        self.test_dataset_parsed = None
        self.train_dataset_parsed = None
        self.val_dataset_parsed = None

        self.train_dataloader_length = 0
        self.train_epoch_step = 0

        self.val_dataloader_names = []
        self.val_dataloader_lengths = []

        self.test_dataloader_names = []

    def setup(self, stage):
        self._load_dataset_paths()

        if stage == "fit":
            assert (
                self.train_dataset is not None
            ), "You need to provide a value for --data.train_dataset"
            assert (
                self.val_dataset is not None
            ), "You need to provide a value for --data.val_dataset"

            if self.train_dataset is None:
                self.train_dataset = "chairs-train"
                logger.warning(
                    "--data.train_dataset is not set. It will be set as {}",
                    self.train_dataset,
                )
            if self.train_batch_size is None:
                self.train_batch_size = 8
                logger.warning(
                    "--data.train_batch_size is not set. It will be set to {}",
                    self.train_batch_size,
                )

            self.train_dataset_parsed = self._parse_dataset_selection(
                self.train_dataset
            )
            self.val_dataset_parsed = self._parse_dataset_selection(self.val_dataset)
        elif stage == "predict":
            assert (
                self.predict_dataset is not None
            ), "You need to provide a value for --data.predict_dataset"
            self.parsed_predict_dataset_parsed = self._parse_dataset_selection(
                self.predict_dataset
            )
        elif stage == "test":
            assert (
                self.test_dataset is not None
            ), "You need to provide a value for --data.test_dataset"
            self.test_dataset_parsed = self._parse_dataset_selection(self.test_dataset)
        elif stage == "validate":
            assert (
                self.val_dataset is not None
            ), "You need to provide a value for --data.val_dataset"
            self.val_dataset_parsed = self._parse_dataset_selection(self.val_dataset)

    def predict_dataloader(self):
        return super().predict_dataloader()

    def test_dataloader(self):
        dataset_ids = [self.test_dataset]
        if "sintel" in dataset_ids:
            dataset_ids.remove("sintel")
            dataset_ids.extend(["sintel-clean", "sintel-final"])
        elif "spring" in dataset_ids:
            dataset_ids.append("spring-revonly")

        dataloaders = []
        for dataset_id in dataset_ids:
            dataset_id += "-test"
            dataset_tokens = dataset_id.split("-")
            dataset = getattr(self, f"_get_{dataset_tokens[0]}_dataset")(
                False, *dataset_tokens[1:]
            )
            dataloaders.append(
                DataLoader(
                    dataset,
                    1,
                    shuffle=False,
                    num_workers=1,
                    pin_memory=False,
                    drop_last=False,
                )
            )

            self.test_dataloader_names.append(dataset_id)

        return dataloaders

    def train_dataloader(self):
        if self.train_dataset_parsed is not None:
            train_dataset = None
            for parsed_vals in self.train_dataset_parsed:
                multiplier = parsed_vals[0]
                dataset_name = parsed_vals[1]
                dataset = getattr(self, f"_get_{dataset_name}_dataset")(
                    True, *parsed_vals[2:]
                )
                dataset_mult = dataset
                for _ in range(multiplier - 1):
                    dataset_mult += dataset

                if train_dataset is None:
                    train_dataset = dataset_mult
                else:
                    train_dataset += dataset_mult

            train_pin_memory = False if self.train_transform_cuda else True
            train_dataloader = DataLoader(
                train_dataset,
                self.train_batch_size,
                shuffle=True,
                num_workers=self.train_num_workers,
                pin_memory=train_pin_memory,
                drop_last=False,
                persistent_workers=self.train_transform_cuda,
            )
            self.train_dataloader_length = len(train_dataloader)
            return train_dataloader

    def val_dataloader(self):
        dataloaders = []
        self.val_dataloader_names = []
        self.val_dataloader_lengths = []
        for parsed_vals in self.val_dataset_parsed:
            dataset_name = parsed_vals[1]
            dataset = getattr(self, f"_get_{dataset_name}_dataset")(
                False, *parsed_vals[2:]
            )
            dataloaders.append(
                DataLoader(
                    dataset,
                    1,
                    shuffle=False,
                    num_workers=1,
                    pin_memory=False,
                    drop_last=False,
                    persistent_workers=self.train_transform_cuda,
                )
            )

            self.val_dataloader_names.append("-".join(parsed_vals[1:]))
            self.val_dataloader_lengths.append(len(dataset))

        return dataloaders

    def _load_dataset_paths(self):
        with open(self.dataset_config_path, "r") as f:
            dataset_paths = yaml.safe_load(f)
        for name, path in dataset_paths.items():
            if getattr(self, f"{name}_root_dir") is None:
                setattr(self, f"{name}_root_dir", path)

    def _parse_dataset_selection(
        self,
        dataset_selection: str,
    ) -> list[tuple[str, int]]:
        """Parse the input string into the selected dataset and their multipliers and parameters.

        For example, 'chairs-train+3*sintel-clean-trainval+kitti-2012-train*5' will be parsed into
        [(1, 'chairs', 'train'), (3, 'sintel', 'clean', 'trainval'), (5, 'kitti', '2012', 'train')].

        Parameters
        ----------
        dataset_selection : str
            The string defining the dataset selection. Each dataset is separated by a '+' sign. The multiplier must be either
            in the beginning or the end of one dataset string, connected to a '*' sign. The remaining content must be separated
            by '-' symbols.

        Returns
        -------
        List[Tuple[str, int]]
            The parsed choice of datasets and their number of repetitions.

        Raises
        ------
        ValueError
            If the given string is invalid.
        """
        if dataset_selection is None:
            return []

        dataset_selection = dataset_selection.replace(" ", "")
        datasets = dataset_selection.split("+")
        for i in range(len(datasets)):
            tokens = datasets[i].split("*")
            if len(tokens) == 1:
                datasets[i] = (1,) + tuple(tokens[0].split("-"))
            elif len(tokens) == 2:
                try:
                    mult, params = int(tokens[0]), tokens[1]
                except ValueError:
                    params, mult = tokens[0], int(
                        tokens[1]
                    )  # if the multiplier comes last.
                datasets[i] = (mult,) + tuple(params.split("-"))
            else:
                raise ValueError(
                    "The specified dataset string {:} is invalid. Check the BaseModel.parse_dataset_selection() documentation "
                    "to see how to write a valid selection string."
                )
        return datasets

    def _get_model_output_stride(self):
        if hasattr(self, "trainer") and self.trainer is not None:
            if hasattr(self.trainer.model, "module"):
                return self.trainer.model.module.output_stride
            else:
                return self.trainer.model.output_stride
        else:
            return 1

    ###########################################################################
    # _get_datasets
    ###########################################################################

    def _get_autoflow_dataset(self, is_train: bool, *args: str) -> Dataset:
        device = "cuda" if self.train_transform_cuda else "cpu"
        md = make_divisible

        fbocc_transform = False
        for v in args:
            if v == "fbocc":
                fbocc_transform = True
            else:
                raise ValueError(f"Invalid arg: {v}")

        if is_train:
            if self.train_crop_size is None:
                cy, cx = (
                    md(368, self._get_model_output_stride()),
                    md(496, self._get_model_output_stride()),
                )
                self.train_crop_size = (cy, cx)
                logger.warning(
                    "--train_crop_size is not set. It will be set as ({}, {}}).", cy, cx
                )
            else:
                cy, cx = (
                    md(self.train_crop_size[0], self._get_model_output_stride()),
                    md(self.train_crop_size[1], self._get_model_output_stride()),
                )

            # These transforms are based on RAFT: https://github.com/princeton-vl/RAFT
            transform = ft.Compose(
                [
                    ft.ToTensor(device=device, fp16=self.train_transform_fp16),
                    ft.RandomScaleAndCrop(
                        (cy, cx),
                        (-0.1, 1.0),
                        (-0.2, 0.2),
                    ),
                    ft.ColorJitter(0.4, 0.4, 0.4, 0.5 / 3.14, 0.2),
                    ft.GaussianNoise(0.02),
                    ft.RandomPatchEraser(
                        0.5, (int(1), int(3)), (int(50), int(100)), "mean"
                    ),
                    ft.RandomFlip(min(0.5, 0.5), min(0.1, 0.5)),
                    (
                        ft.GenerateFBCheckFlowOcclusion(threshold=1)
                        if fbocc_transform
                        else None
                    ),
                ]
            )
        else:
            transform = ft.ToTensor()

        split = "trainval"
        if len(args) > 0 and args[0] in ["train", "val", "trainval"]:
            split = args[0]
        dataset = AutoFlowDataset(
            self.autoflow_root_dir, split=split, transform=transform
        )
        return dataset

    def _get_chairs_dataset(self, is_train: bool, *args: str) -> Dataset:
        device = "cuda" if self.train_transform_cuda else "cpu"
        md = make_divisible

        fbocc_transform = False
        split = "trainval"
        for v in args:
            if v in ["train", "val", "trainval"]:
                split = args[0]
            elif v == "fbocc":
                fbocc_transform = True
            else:
                raise ValueError(f"Invalid arg: {v}")

        if is_train:
            if self.train_crop_size is None:
                cy, cx = (
                    md(368, self._get_model_output_stride()),
                    md(496, self._get_model_output_stride()),
                )
                self.train_crop_size = (cy, cx)
                logger.warning(
                    "--train_crop_size is not set. It will be set as ({}, {}).", cy, cx
                )
            else:
                cy, cx = (
                    md(self.train_crop_size[0], self._get_model_output_stride()),
                    md(self.train_crop_size[1], self._get_model_output_stride()),
                )

            # These transforms are based on RAFT: https://github.com/princeton-vl/RAFT
            transform = ft.Compose(
                [
                    ft.ToTensor(device=device, fp16=self.train_transform_fp16),
                    ft.RandomScaleAndCrop(
                        (cy, cx),
                        (-0.1, 1.0),
                        (-0.2, 0.2),
                    ),
                    ft.ColorJitter(0.4, 0.4, 0.4, 0.5 / 3.14, 0.2),
                    ft.GaussianNoise(0.02),
                    ft.RandomPatchEraser(
                        0.5, (int(1), int(3)), (int(50), int(100)), "mean"
                    ),
                    ft.RandomFlip(min(0.5, 0.5), min(0.1, 0.5)),
                    (
                        ft.GenerateFBCheckFlowOcclusion(threshold=1)
                        if fbocc_transform
                        else None
                    ),
                ]
            )
        else:
            transform = ft.ToTensor()

        dataset = FlyingChairsDataset(
            self.flying_chairs_root_dir, split=split, transform=transform
        )
        return dataset

    def _get_chairs2_dataset(self, is_train: bool, *args: str) -> Dataset:
        device = "cuda" if self.train_transform_cuda else "cpu"
        md = make_divisible

        split = "trainval"
        add_reverse = False
        get_occlusion_mask = False
        get_motion_boundary_mask = False
        get_backward = False
        fbocc_transform = False
        for v in args:
            if v in ["train", "val", "trainval"]:
                split = v
            elif v == "rev":
                add_reverse = True
            elif v == "occ":
                get_occlusion_mask = True
            elif v == "mb":
                get_motion_boundary_mask = True
            elif v == "back":
                get_backward = True
            elif v == "fbocc":
                fbocc_transform = True
            else:
                raise ValueError(f"Invalid arg: {v}")

        if is_train:
            if self.train_crop_size is None:
                cy, cx = (
                    md(368, self._get_model_output_stride()),
                    md(496, self._get_model_output_stride()),
                )
                self.train_crop_size = (cy, cx)
                logger.warning(
                    "--train_crop_size is not set. It will be set as ({}, {}).", cy, cx
                )
            else:
                cy, cx = (
                    md(self.train_crop_size[0], self._get_model_output_stride()),
                    md(self.train_crop_size[1], self._get_model_output_stride()),
                )

            # These transforms are based on RAFT: https://github.com/princeton-vl/RAFT
            transform = ft.Compose(
                [
                    ft.ToTensor(device=device, fp16=self.train_transform_fp16),
                    ft.RandomScaleAndCrop(
                        (cy, cx),
                        (-0.1, 1.0),
                        (-0.2, 0.2),
                    ),
                    ft.ColorJitter(0.4, 0.4, 0.4, 0.5 / 3.14, 0.2),
                    ft.GaussianNoise(0.02),
                    ft.RandomPatchEraser(
                        0.5, (int(1), int(3)), (int(50), int(100)), "mean"
                    ),
                    ft.RandomFlip(min(0.5, 0.5), min(0.1, 0.5)),
                    (
                        ft.GenerateFBCheckFlowOcclusion(threshold=1)
                        if fbocc_transform
                        else None
                    ),
                ]
            )
        else:
            transform = ft.ToTensor()

        dataset = FlyingChairs2Dataset(
            self.flying_chairs2_root_dir,
            split=split,
            transform=transform,
            add_reverse=add_reverse,
            get_occlusion_mask=get_occlusion_mask,
            get_motion_boundary_mask=get_motion_boundary_mask,
            get_backward=get_backward,
        )
        return dataset

    def _get_hd1k_dataset(self, is_train: bool, *args: str) -> Dataset:
        device = "cuda" if self.train_transform_cuda else "cpu"
        md = make_divisible

        split = "trainval"
        sequence_length = 2
        sequence_position = "first"
        fbocc_transform = False
        for v in args:
            if v in ["train", "val", "trainval", "test"]:
                split = args[0]
            elif v.startswith("seqlen"):
                sequence_length = int(v.split("_")[1])
            elif v.startswith("seqpos"):
                sequence_position = v.split("_")[1]
            elif v == "fbocc":
                fbocc_transform = True
            else:
                raise ValueError(f"Invalid arg: {v}")

        if is_train:
            if self.train_crop_size is None:
                cy, cx = (
                    md(368, self._get_model_output_stride()),
                    md(768, self._get_model_output_stride()),
                )
                self.train_crop_size = (cy, cx)
                logger.warning(
                    "--train_crop_size is not set. It will be set as ({}, {}).", cy, cx
                )
            else:
                cy, cx = (
                    md(self.train_crop_size[0], self._get_model_output_stride()),
                    md(self.train_crop_size[1], self._get_model_output_stride()),
                )

            # These transforms are based on RAFT: https://github.com/princeton-vl/RAFT
            transform = ft.Compose(
                [
                    ft.ToTensor(device=device, fp16=self.train_transform_fp16),
                    ft.RandomScaleAndCrop(
                        (cy, cx), (-0.5, 0.2), (-0.2, 0.2), sparse=True
                    ),
                    ft.ColorJitter(0.4, 0.4, 0.4, 0.5 / 3.14, 0.2),
                    ft.GaussianNoise(0.02),
                    ft.RandomPatchEraser(
                        0.5, (int(1), int(3)), (int(50), int(100)), "mean"
                    ),
                    ft.RandomFlip(min(0.5, 0.5), min(0.1, 0.5)),
                    (
                        ft.GenerateFBCheckFlowOcclusion(threshold=1)
                        if fbocc_transform
                        else None
                    ),
                ]
            )
        else:
            transform = ft.ToTensor()

        dataset = Hd1kDataset(
            self.hd1k_root_dir,
            split=split,
            transform=transform,
            sequence_length=sequence_length,
            sequence_position=sequence_position,
        )
        return dataset

    def _get_kitti_dataset(self, is_train: bool, *args: str) -> Dataset:
        device = "cuda" if self.train_transform_cuda else "cpu"
        md = make_divisible

        versions = ["2012", "2015"]
        split = "trainval"
        get_occlusion_mask = False
        fbocc_transform = False
        for v in args:
            if v in ["2012", "2015"]:
                versions = [v]
            elif v in ["train", "val", "trainval", "test"]:
                split = v
            elif v == "occ":
                get_occlusion_mask = True
            elif v == "fbocc":
                fbocc_transform = True
            else:
                raise ValueError(f"Invalid arg: {v}")

        if is_train:
            if self.train_crop_size is None:
                cy, cx = (
                    md(288, self._get_model_output_stride()),
                    md(960, self._get_model_output_stride()),
                )
                # cy, cx = (md(416, self._get_model_output_stride()), md(960, self._get_model_output_stride()))
                self.train_crop_size = (cy, cx)
                logger.warning(
                    "--train_crop_size is not set. It will be set as ({}, {}).", cy, cx
                )
            else:
                cy, cx = (
                    md(self.train_crop_size[0], self._get_model_output_stride()),
                    md(self.train_crop_size[1], self._get_model_output_stride()),
                )

            # These transforms are based on RAFT: https://github.com/princeton-vl/RAFT
            transform = ft.Compose(
                [
                    ft.ToTensor(device=device, fp16=self.train_transform_fp16),
                    ft.RandomScaleAndCrop(
                        (cy, cx), (-0.2, 0.4), (-0.2, 0.2), sparse=True
                    ),
                    ft.ColorJitter(0.4, 0.4, 0.4, 0.5 / 3.14, 0.2),
                    ft.GaussianNoise(0.02),
                    ft.RandomPatchEraser(
                        0.5, (int(1), int(3)), (int(50), int(100)), "mean"
                    ),
                    (
                        ft.GenerateFBCheckFlowOcclusion(threshold=1)
                        if fbocc_transform
                        else None
                    ),
                ]
            )
        else:
            transform = ft.ToTensor()

        dataset = KittiDataset(
            self.kitti_2012_root_dir,
            self.kitti_2015_root_dir,
            get_occlusion_mask=get_occlusion_mask,
            versions=versions,
            split=split,
            transform=transform,
        )
        return dataset

    def _get_kubric_dataset(self, is_train: bool, *args: str) -> Dataset:
        if is_train:
            raise NotImplementedError()
        else:
            transform = ft.ToTensor()

        get_backward = False
        sequence_length = 2
        sequence_position = "first"
        max_seq = None
        for v in args:
            if v == "back":
                get_backward = True
            elif v.startswith("seqlen"):
                sequence_length = int(v.split("_")[1])
            elif v.startswith("seqpos"):
                sequence_position = v.split("_")[1]
            elif v.startswith("maxseq"):
                max_seq = int(v.split("_")[1])

        dataset = KubricDataset(
            self.kubric_root_dir,
            transform=transform,
            get_backward=get_backward,
            sequence_length=sequence_length,
            sequence_position=sequence_position,
            max_seq=max_seq,
        )
        return dataset

    def _get_sintel_dataset(self, is_train: bool, *args: str) -> Dataset:
        device = "cuda" if self.train_transform_cuda else "cpu"
        md = make_divisible

        pass_names = ["clean", "final"]
        split = "trainval"
        get_occlusion_mask = False
        sequence_length = 2
        sequence_position = "first"
        fbocc_transform = False
        for v in args:
            if v in ["clean", "final"]:
                pass_names = [v]
            elif v in ["train", "val", "trainval", "test"]:
                split = v
            elif v == "occ":
                get_occlusion_mask = True
            elif v.startswith("seqlen"):
                sequence_length = int(v.split("_")[1])
            elif v.startswith("seqpos"):
                sequence_position = v.split("_")[1]
            elif v == "fbocc":
                fbocc_transform = True
            else:
                raise ValueError(f"Invalid arg: {v}")

        if is_train:
            if self.train_crop_size is None:
                cy, cx = (
                    md(368, self._get_model_output_stride()),
                    md(768, self._get_model_output_stride()),
                )
                self.train_crop_size = (cy, cx)
                logger.warning(
                    "--train_crop_size is not set. It will be set as ({}, {}).", cy, cx
                )
            else:
                cy, cx = (
                    md(self.train_crop_size[0], self._get_model_output_stride()),
                    md(self.train_crop_size[1], self._get_model_output_stride()),
                )

            # These transforms are based on RAFT: https://github.com/princeton-vl/RAFT
            transform = ft.Compose(
                [
                    ft.ToTensor(device=device, fp16=self.train_transform_fp16),
                    ft.RandomScaleAndCrop((cy, cx), (-0.2, 0.6), (-0.2, 0.2)),
                    ft.ColorJitter(0.4, 0.4, 0.4, 0.5 / 3.14, 0.2),
                    ft.GaussianNoise(0.02),
                    ft.RandomPatchEraser(
                        0.5, (int(1), int(3)), (int(50), int(100)), "mean"
                    ),
                    ft.RandomFlip(min(0.5, 0.5), min(0.1, 0.5)),
                    (
                        ft.GenerateFBCheckFlowOcclusion(threshold=1)
                        if fbocc_transform
                        else None
                    ),
                ]
            )
        else:
            transform = ft.ToTensor()

        dataset = SintelDataset(
            self.mpi_sintel_root_dir,
            split=split,
            pass_names=pass_names,
            transform=transform,
            get_occlusion_mask=get_occlusion_mask,
            sequence_length=sequence_length,
            sequence_position=sequence_position,
        )
        return dataset

    def _get_sintel_finetune_dataset(self, is_train: bool, *args: str) -> Dataset:
        device = "cuda" if self.train_transform_cuda else "cpu"
        md = make_divisible

        fbocc_transform = False
        searaft_split = False
        for v in args:
            if v == "fbocc":
                fbocc_transform = True
            elif v == "searaft_split":
                searaft_split = True
            else:
                raise ValueError(f"Invalid arg: {v}")

        if is_train:
            if self.train_crop_size is None:
                cy, cx = (
                    md(368, self._get_model_output_stride()),
                    md(768, self._get_model_output_stride()),
                )
                # cy, cx = (md(416, self._get_model_output_stride()), md(960, self._get_model_output_stride()))
                self.train_crop_size = (cy, cx)
                logger.warning(
                    "--train_crop_size is not set. It will be set as ({}, {}).", cy, cx
                )
            else:
                cy, cx = (
                    md(self.train_crop_size[0], self._get_model_output_stride()),
                    md(self.train_crop_size[1], self._get_model_output_stride()),
                )

            # These transforms are based on RAFT: https://github.com/princeton-vl/RAFT
            transform1 = ft.Compose(
                [
                    ft.ToTensor(device=device, fp16=self.train_transform_fp16),
                    ft.RandomScaleAndCrop((cy, cx), (-0.2, 0.6), (-0.2, 0.2)),
                    ft.ColorJitter(0.4, 0.4, 0.4, 0.5 / 3.14, 0.2),
                    ft.RandomPatchEraser(
                        0.5, (int(1), int(3)), (int(50), int(100)), "mean"
                    ),
                    ft.RandomFlip(min(0.5, 0.5), min(0.1, 0.5)),
                    (
                        ft.GenerateFBCheckFlowOcclusion(threshold=1)
                        if fbocc_transform
                        else None
                    ),
                ]
            )
            things_dataset = FlyingThings3DDataset(
                self.flying_things3d_root_dir,
                split="train",
                pass_names=["clean"],
                side_names=["left"],
                transform=transform1,
                get_backward=False,
                get_motion_boundary_mask=False,
                get_occlusion_mask=False,
            )

            sintel_clean_dataset = SintelDataset(
                self.mpi_sintel_root_dir,
                split="trainval",
                pass_names=["clean"],
                transform=transform1,
                get_occlusion_mask=False,
            )
            sintel_clean_mult_dataset = sintel_clean_dataset
            for _ in range(19 if searaft_split else 99):
                sintel_clean_mult_dataset += sintel_clean_dataset

            sintel_final_dataset = SintelDataset(
                self.mpi_sintel_root_dir,
                split="trainval",
                pass_names=["final"],
                transform=transform1,
                get_occlusion_mask=False,
            )
            sintel_final_mult_dataset = sintel_final_dataset
            for _ in range(19 if searaft_split else 99):
                sintel_final_mult_dataset += sintel_final_dataset

            transform2 = ft.Compose(
                [
                    ft.ToTensor(device=device, fp16=self.train_transform_fp16),
                    ft.RandomScaleAndCrop(
                        (cy, cx), (-0.3, 0.5), (-0.2, 0.2), sparse=True
                    ),
                    ft.ColorJitter(0.4, 0.4, 0.4, 0.5 / 3.14, 0.2),
                    ft.RandomPatchEraser(
                        0.5, (int(1), int(3)), (int(50), int(100)), "mean"
                    ),
                    ft.RandomFlip(min(0.5, 0.5), min(0.1, 0.5)),
                    (
                        ft.GenerateFBCheckFlowOcclusion(threshold=1)
                        if fbocc_transform
                        else None
                    ),
                ]
            )
            kitti_dataset = KittiDataset(
                root_dir_2015=self.kitti_2015_root_dir,
                split="trainval",
                versions=["2015"],
                transform=transform2,
                get_occlusion_mask=False,
            )
            kitti_mult_dataset = kitti_dataset
            for _ in range(79 if searaft_split else 199):
                kitti_mult_dataset += kitti_dataset

            transform3 = ft.Compose(
                [
                    ft.ToTensor(device=device, fp16=self.train_transform_fp16),
                    ft.RandomScaleAndCrop(
                        (cy, cx), (-0.5, 0.2), (-0.2, 0.2), sparse=True
                    ),
                    ft.ColorJitter(0.4, 0.4, 0.4, 0.5 / 3.14, 0.2),
                    ft.RandomPatchEraser(
                        0.5, (int(1), int(3)), (int(50), int(100)), "mean"
                    ),
                    ft.RandomFlip(min(0.5, 0.5), min(0.1, 0.5)),
                    (
                        ft.GenerateFBCheckFlowOcclusion(threshold=1)
                        if fbocc_transform
                        else None
                    ),
                ]
            )
            hd1k_dataset = Hd1kDataset(
                self.hd1k_root_dir, split="trainval", transform=transform3
            )
            hd1k_mult_dataset = hd1k_dataset
            for _ in range(29 if searaft_split else 4):
                hd1k_mult_dataset += hd1k_dataset

            mixed_dataset = (
                things_dataset
                + sintel_clean_mult_dataset
                + sintel_final_mult_dataset
                + kitti_mult_dataset
                + hd1k_mult_dataset
            )

            logger.info("Loaded datasets:")
            logger.info(
                "FlyingThings3D: unique samples {} - multiplied samples {}",
                len(things_dataset),
                len(things_dataset),
            )
            logger.info(
                "Sintel clean: unique samples {} - multiplied samples {}",
                len(sintel_clean_dataset),
                len(sintel_clean_mult_dataset),
            )
            logger.info(
                "Sintel final: unique samples {} - multiplied samples {}",
                len(sintel_final_dataset),
                len(sintel_final_mult_dataset),
            )
            logger.info(
                "KITTI 2015: unique samples {} - multiplied samples {}",
                len(kitti_dataset),
                len(kitti_mult_dataset),
            )
            logger.info(
                "HD1K: unique samples {} - multiplied samples {}",
                len(hd1k_dataset),
                len(hd1k_mult_dataset),
            )
            logger.info("Total dataset size: {}", len(mixed_dataset))
        else:
            raise NotImplementedError()

        return mixed_dataset

    def _get_spring_dataset(self, is_train: bool, *args: str) -> Dataset:
        device = "cuda" if self.train_transform_cuda else "cpu"
        md = make_divisible

        split = "train"
        add_reverse = False
        get_backward = False
        sequence_length = 2
        sequence_position = "first"
        reverse_only = False
        subsample = False
        side_names = []
        fbocc_transform = False
        for v in args:
            if v in ["train", "val", "trainval", "test"]:
                split = v
            elif v == "rev":
                add_reverse = True
            elif v == "revonly":
                reverse_only = True
            elif v == "back":
                get_backward = True
            elif v.startswith("seqlen"):
                sequence_length = int(v.split("_")[1])
            elif v.startswith("seqpos"):
                sequence_position = v.split("_")[1]
            elif v == "sub":
                subsample = True
            elif v == "left":
                side_names.append("left")
            elif v == "right":
                side_names.append("right")
            elif v == "fbocc":
                fbocc_transform = True
            else:
                raise ValueError(f"Invalid arg: {v}")

        if is_train:
            if self.train_crop_size is None:
                cy, cx = (
                    md(540, self._get_model_output_stride()),
                    md(960, self._get_model_output_stride()),
                )
                self.train_crop_size = (cy, cx)
                logger.warning(
                    "--train_crop_size is not set. It will be set as ({}, {}).", cy, cx
                )
            else:
                cy, cx = (
                    md(self.train_crop_size[0], self._get_model_output_stride()),
                    md(self.train_crop_size[1], self._get_model_output_stride()),
                )

            # Transforms copied from SEA-RAFT
            transform = ft.Compose(
                [
                    ft.ToTensor(device=device, fp16=self.train_transform_fp16),
                    ft.RandomScaleAndCrop((cy, cx), (0.0, 0.2), (-0.2, 0.2)),
                    ft.ColorJitter(0.4, 0.4, 0.4, 0.5 / 3.14, 0.2),
                    ft.GaussianNoise(0.02),
                    ft.RandomPatchEraser(
                        0.5, (int(1), int(3)), (int(50), int(100)), "mean"
                    ),
                    ft.RandomFlip(min(0.5, 0.5), min(0.1, 0.5)),
                    (
                        ft.GenerateFBCheckFlowOcclusion(threshold=1)
                        if fbocc_transform
                        else None
                    ),
                ]
            )
        else:
            transform = ft.ToTensor()

        if len(side_names) == 0:
            side_names = ["left", "right"]

        dataset = SpringDataset(
            self.spring_root_dir,
            split=split,
            side_names=side_names,
            add_reverse=add_reverse,
            transform=transform,
            get_backward=get_backward,
            sequence_length=sequence_length,
            sequence_position=sequence_position,
            reverse_only=reverse_only,
            subsample=subsample,
        )
        return dataset

    def _get_tartanair_dataset(self, is_train: bool, *args: str) -> Dataset:
        device = "cuda" if self.train_transform_cuda else "cpu"
        md = make_divisible

        get_occlusion_mask = False
        sequence_length = 2
        sequence_position = "first"
        difficulties = []
        fbocc_transform = False
        for v in args:
            if v in ["easy", "hard"]:
                difficulties.append(v)
            elif v == "occ":
                get_occlusion_mask = True
            elif v.startswith("seqlen"):
                sequence_length = int(v.split("_")[1])
            elif v.startswith("seqpos"):
                sequence_position = v.split("_")[1]
            elif v == "fbocc":
                fbocc_transform = True
            else:
                raise ValueError(f"Invalid arg: {v}")

        if len(difficulties) == 0:
            difficulties = ["easy"]

        if is_train:
            if self.train_crop_size is None:
                cy, cx = (
                    md(360, self._get_model_output_stride()),
                    md(480, self._get_model_output_stride()),
                )
                self.train_crop_size = (cy, cx)
                logger.warning(
                    "--train_crop_size is not set. It will be set as ({}, {}).", cy, cx
                )
            else:
                cy, cx = (
                    md(self.train_crop_size[0], self._get_model_output_stride()),
                    md(self.train_crop_size[1], self._get_model_output_stride()),
                )

            # These transforms are based on RAFT: https://github.com/princeton-vl/RAFT
            transform = ft.Compose(
                [
                    ft.ToTensor(device=device, fp16=self.train_transform_fp16),
                    ft.RandomScaleAndCrop((cy, cx), (-0.4, 0.8), (-0.2, 0.2)),
                    ft.ColorJitter(0.4, 0.4, 0.4, 0.5 / 3.14, 0.2),
                    ft.GaussianNoise(0.02),
                    ft.RandomPatchEraser(
                        0.5, (int(1), int(3)), (int(50), int(100)), "mean"
                    ),
                    ft.RandomFlip(min(0.5, 0.5), min(0.1, 0.5)),
                    (
                        ft.GenerateFBCheckFlowOcclusion(threshold=1)
                        if fbocc_transform
                        else None
                    ),
                ]
            )
        else:
            transform = ft.ToTensor()

        dataset = TartanAirDataset(
            self.tartanair_root_dir,
            difficulties=difficulties,
            transform=transform,
            get_occlusion_mask=get_occlusion_mask,
            sequence_length=sequence_length,
            sequence_position=sequence_position,
        )
        return dataset

    def _get_things_dataset(self, is_train: bool, *args: str) -> Dataset:
        device = "cuda" if self.train_transform_cuda else "cpu"
        md = make_divisible

        pass_names = ["clean", "final"]
        split = "trainval"
        is_subset = False
        add_reverse = False
        get_occlusion_mask = False
        get_motion_boundary_mask = False
        get_backward = False
        sequence_length = 2
        sequence_position = "first"
        sintel_transform = False
        fbocc_transform = False
        for v in args:
            if v in ["clean", "final"]:
                pass_names = [v]
            elif v in ["train", "val", "trainval"]:
                split = v
            elif v == "subset":
                is_subset = True
            elif v == "rev":
                add_reverse = True
            elif v == "occ":
                get_occlusion_mask = True
            elif v == "mb":
                get_motion_boundary_mask = True
            elif v == "back":
                get_backward = True
            elif v.startswith("seqlen"):
                sequence_length = int(v.split("_")[1])
            elif v.startswith("seqpos"):
                sequence_position = v.split("_")[1]
            elif v == "sinteltransform":
                sintel_transform = True
            elif v == "fbocc":
                fbocc_transform = True
            else:
                raise ValueError(f"Invalid arg: {v}")

        if is_train:
            if self.train_crop_size is None:
                cy, cx = (
                    md(400, self._get_model_output_stride()),
                    md(720, self._get_model_output_stride()),
                )
                # cy, cx = (md(416, self._get_model_output_stride()), md(960, self._get_model_output_stride()))
                self.train_crop_size = (cy, cx)
                logger.warning(
                    "--train_crop_size is not set. It will be set as ({}, {}).", cy, cx
                )
            else:
                cy, cx = (
                    md(self.train_crop_size[0], self._get_model_output_stride()),
                    md(self.train_crop_size[1], self._get_model_output_stride()),
                )

            # These transforms are based on RAFT: https://github.com/princeton-vl/RAFT
            if sintel_transform:
                major_scale = (-0.2, 0.6)
            else:
                major_scale = (-0.4, 0.8)
            transform = ft.Compose(
                [
                    ft.ToTensor(device=device, fp16=self.train_transform_fp16),
                    ft.RandomScaleAndCrop((cy, cx), major_scale, (-0.2, 0.2)),
                    ft.ColorJitter(0.4, 0.4, 0.4, 0.5 / 3.14, 0.2),
                    ft.GaussianNoise(0.02),
                    ft.RandomPatchEraser(
                        0.5, (int(1), int(3)), (int(50), int(100)), "mean"
                    ),
                    ft.RandomFlip(min(0.5, 0.5), min(0.1, 0.5)),
                    (
                        ft.GenerateFBCheckFlowOcclusion(threshold=1)
                        if fbocc_transform
                        else None
                    ),
                ]
            )
        else:
            transform = ft.ToTensor()

        if is_subset:
            dataset = FlyingThings3DSubsetDataset(
                self.flying_things3d_subset_root_dir,
                split=split,
                pass_names=pass_names,
                side_names=["left", "right"],
                add_reverse=add_reverse,
                transform=transform,
                get_occlusion_mask=get_occlusion_mask,
                get_motion_boundary_mask=get_motion_boundary_mask,
                get_backward=get_backward,
                sequence_length=sequence_length,
                sequence_position=sequence_position,
            )
        else:
            dataset = FlyingThings3DDataset(
                self.flying_things3d_root_dir,
                split=split,
                pass_names=pass_names,
                side_names=["left", "right"],
                add_reverse=add_reverse,
                transform=transform,
                get_occlusion_mask=get_occlusion_mask,
                get_motion_boundary_mask=get_motion_boundary_mask,
                get_backward=get_backward,
                sequence_length=sequence_length,
                sequence_position=sequence_position,
            )
        return dataset

    def _get_middlebury_st_dataset(self, is_train: bool, *args: str) -> Dataset:
        assert not is_train
        transform = ft.ToTensor()

        dataset = MiddleburySTDataset(
            self.middlebury_st_root_dir,
            transform=transform,
        )
        return dataset

    def _get_viper_dataset(self, is_train: bool, *args: str) -> Dataset:
        assert not is_train
        transform = ft.ToTensor()

        dataset = ViperDataset(
            self.viper_root_dir,
            split="val",
            transform=transform,
        )
        return dataset

    def _get_overfit_dataset(self, is_train: bool, *args: str) -> Dataset:
        md = make_divisible
        if self.train_crop_size is None:
            cy, cx = (
                md(436, self._get_model_output_stride()),
                md(1024, self._get_model_output_stride()),
            )
            self.train_crop_size = (cy, cx)
            logger.warning(
                "--train_crop_size is not set. It will be set as ({}, {}).", cy, cx
            )
        else:
            cy, cx = (
                md(self.train_crop_size[0], self._get_model_output_stride()),
                md(self.train_crop_size[1], self._get_model_output_stride()),
            )
        transform = ft.Compose([ft.ToTensor(), ft.Resize((cy, cx))])

        dataset_name = "sintel"
        if len(args) > 0 and args[0] in ["chairs2"]:
            dataset_name = args[0]

        if dataset_name == "sintel":
            dataset = SintelDataset(
                self.mpi_sintel_root_dir,
                split="trainval",
                pass_names="clean",
                transform=transform,
                get_occlusion_mask=False,
            )
        elif dataset_name == "chairs2":
            dataset = FlyingChairs2Dataset(
                self.flying_chairs2_root_dir,
                split="trainval",
                transform=transform,
                add_reverse=False,
                get_occlusion_mask=True,
                get_motion_boundary_mask=True,
                get_backward=True,
            )

        dataset.img_paths = dataset.img_paths[:1]
        dataset.flow_paths = dataset.flow_paths[:1]
        dataset.occ_paths = dataset.occ_paths[:1]
        dataset.mb_paths = dataset.mb_paths[:1]
        dataset.flow_b_paths = dataset.flow_b_paths[:1]
        dataset.occ_b_paths = dataset.occ_b_paths[:1]
        dataset.mb_b_paths = dataset.mb_b_paths[:1]
        dataset.metadata = dataset.metadata[:1]

        return dataset



"""
Portions of this code copyright 2017, Clement Pinard
"""

# freda (todo) : adversarial loss

import torch
import torch.nn as nn
import math


def EPE(input_flow, target_flow):
    return torch.norm(target_flow - input_flow, p=2, dim=1).mean()


class L1(nn.Module):
    def __init__(self):
        super(L1, self).__init__()

    def forward(self, output, target):
        lossvalue = torch.abs(output - target).mean()
        return lossvalue


class L2(nn.Module):
    def __init__(self):
        super(L2, self).__init__()

    def forward(self, output, target):
        lossvalue = torch.norm(output - target, p=2, dim=1).mean()
        return lossvalue


class L1Loss(nn.Module):
    def __init__(self):
        super(L1Loss, self).__init__()
        self.loss = L1()
        self.loss_labels = ["L1", "EPE"]

    def forward(self, output, target):
        output = output["flow_preds"]
        target = target["flows"][:, 0]
        lossvalue = self.loss(output, target)
        return lossvalue


class L2Loss(nn.Module):
    def __init__(self):
        super(L2Loss, self).__init__()
        self.loss = L2()
        self.loss_labels = ["L2", "EPE"]

    def forward(self, output, target):
        output = output["flow_preds"]
        target = target["flows"][:, 0]
        lossvalue = self.loss(output, target)
        return lossvalue


class MultiScale(nn.Module):
    def __init__(self, startScale=4, numScales=5, l_weight=0.32, norm="L1"):
        super(MultiScale, self).__init__()

        self.startScale = startScale
        self.numScales = numScales
        self.loss_weights = torch.FloatTensor(
            [(l_weight / 2**scale) for scale in range(self.numScales)]
        )
        self.l_type = norm
        self.div_flow = 0.05
        assert len(self.loss_weights) == self.numScales

        if self.l_type == "L1":
            self.loss = L1()
        else:
            self.loss = L2()

        self.multiScales = [
            nn.AvgPool2d(self.startScale * (2**scale), self.startScale * (2**scale))
            for scale in range(self.numScales)
        ]
        self.loss_labels = ["MultiScale-" + self.l_type, "EPE"]

    def forward(self, output, target):
        output = output["flow_preds"]
        target = target["flows"][:, 0]
        lossvalue = 0

        if type(output) is tuple or type(output) is list:
            target = self.div_flow * target
            for i, flow_pred_ in enumerate(output):
                target_ = self.multiScales[i](target)
                lossvalue += self.loss_weights[i] * self.loss(flow_pred_, target_)
        else:
            lossvalue += self.loss(output, target)
        return lossvalue



import torch
import torch.nn.functional as F
import torch.nn as nn
from .utils import bilinear_sampler

try:
    from .libs.GANet.modules.GANet import NLFIter
except (ModuleNotFoundError, ImportError):
    NLFIter = None

try:
    import alt_cuda_corr
except:
    # alt_cuda_corr is not compiled
    pass


class NLF(nn.Module):
    def __init__(self, in_channel=32):
        super(NLF, self).__init__()
        self.nlf = NLFIter()

    def forward(self, x, g):
        N, D1, D2, H, W = x.shape
        x = x.reshape(N, D1 * D2, H, W).contiguous()
        rem = x
        k1, k2, k3, k4 = torch.split(g, (5, 5, 5, 5), 1)
        #        k1, k2, k3, k4 = self.getweights(x)
        k1 = F.normalize(k1, p=1, dim=1)
        k2 = F.normalize(k2, p=1, dim=1)
        k3 = F.normalize(k3, p=1, dim=1)
        k4 = F.normalize(k4, p=1, dim=1)

        x = self.nlf(x, k1, k2, k3, k4)
        #        x = x + rem
        x = x.reshape(N, D1, D2, H, W)
        return x


class CorrBlock:
    def __init__(self, fmap1, fmap2, guid, num_levels=4, radius=4):
        self.num_levels = num_levels
        self.radius = radius
        self.corr_pyramid = []

        # all pairs correlation
        self.nlf = NLF()
        corr = self.corr_compute(fmap1, fmap2, guid, reverse=True)
        # corr = self.nlf(corr, g)
        # corr = corr.permute(0, 3,4, 1,2)

        batch, h1, w1, h2, w2 = corr.shape
        self.shape = corr.shape
        corr = corr.reshape(batch * h1 * w1, 1, h2, w2)

        self.corr_pyramid.append(corr)
        for i in range(self.num_levels - 1):
            corr = F.avg_pool2d(corr, 2, stride=2)
            self.corr_pyramid.append(corr)

    def separate(self):
        sep_u = []
        sep_v = []
        for i in range(self.num_levels):
            corr = self.corr_pyramid[i]
            m1, _ = corr.max(dim=2, keepdim=True)
            m2 = corr.mean(dim=2, keepdim=True)
            sep = torch.cat((m1, m2), dim=2)
            sep = sep.reshape(
                self.shape[0], self.shape[1], self.shape[2], sep.shape[2], sep.shape[3]
            ).permute(0, 3, 4, 1, 2)
            sep = F.interpolate(
                sep,
                [self.shape[4], self.shape[1], self.shape[2]],
                mode="trilinear",
                align_corners=True,
            )
            sep_u.append(sep)
            m1, _ = corr.max(dim=3, keepdim=True)
            m2 = corr.mean(dim=3, keepdim=True)
            sep = torch.cat((m1, m2), dim=3)
            sep = sep.reshape(
                self.shape[0], self.shape[1], self.shape[2], sep.shape[2], sep.shape[3]
            ).permute(0, 4, 3, 1, 2)
            sep = F.interpolate(
                sep,
                [self.shape[3], self.shape[1], self.shape[2]],
                mode="trilinear",
                align_corners=True,
            )
            sep_v.append(sep)
        sep_u = torch.cat(sep_u, dim=1)
        sep_v = torch.cat(sep_v, dim=1)
        return sep_u, sep_v

    def __call__(self, coords, sep=False):
        if sep:
            return self.separate()
        r = self.radius
        coords = coords.permute(0, 2, 3, 1)
        batch, h1, w1, _ = coords.shape

        out_pyramid = []
        for i in range(self.num_levels):
            corr = self.corr_pyramid[i]
            dx = torch.linspace(-r, r, 2 * r + 1)
            dy = torch.linspace(-r, r, 2 * r + 1)
            delta = torch.stack(torch.meshgrid(dy, dx, indexing="ij"), axis=-1).to(
                coords.device
            )

            centroid_lvl = coords.reshape(batch * h1 * w1, 1, 1, 2) / 2**i
            delta_lvl = delta.view(1, 2 * r + 1, 2 * r + 1, 2)
            coords_lvl = centroid_lvl + delta_lvl

            corr = bilinear_sampler(corr, coords_lvl)
            corr = corr.view(batch, h1, w1, -1)
            out_pyramid.append(corr)

        out = torch.cat(out_pyramid, dim=-1)
        return out.permute(0, 3, 1, 2).contiguous().float()

    # @staticmethod
    def corr_compute(self, fmap1, fmap2, guid, reverse=True):
        batch, dim, ht, wd = fmap1.shape
        fmap1 = fmap1.view(batch, dim, ht * wd)
        fmap2 = fmap2.view(batch, dim, ht * wd)

        if reverse:
            corr = torch.matmul(fmap2.transpose(1, 2), fmap1) / torch.sqrt(
                torch.tensor(dim).float()
            )
            corr = corr.view(batch, ht, wd, ht, wd)
            corr = self.nlf(corr, guid)
            corr = corr.permute(0, 3, 4, 1, 2)
        else:
            corr = torch.matmul(fmap1.transpose(1, 2), fmap2) / torch.sqrt(
                torch.tensor(dim).float()
            )
            corr = corr.view(batch, ht, wd, ht, wd)
            corr = self.nlf(corr, guid)

        return corr


class AlternateCorrBlock:
    def __init__(self, fmap1, fmap2, num_levels=4, radius=4):
        self.num_levels = num_levels
        self.radius = radius

        self.pyramid = [(fmap1, fmap2)]
        for i in range(self.num_levels):
            fmap1 = F.avg_pool2d(fmap1, 2, stride=2)
            fmap2 = F.avg_pool2d(fmap2, 2, stride=2)
            self.pyramid.append((fmap1, fmap2))

    def __call__(self, coords):
        coords = coords.permute(0, 2, 3, 1)
        B, H, W, _ = coords.shape
        dim = self.pyramid[0][0].shape[1]

        corr_list = []
        for i in range(self.num_levels):
            r = self.radius
            fmap1_i = self.pyramid[0][0].permute(0, 2, 3, 1).contiguous()
            fmap2_i = self.pyramid[i][1].permute(0, 2, 3, 1).contiguous()

            coords_i = (coords / 2**i).reshape(B, 1, H, W, 2).contiguous()
            if coords.dtype == torch.float16:
                fmap1_i = fmap1_i.float()
                fmap2_i = fmap2_i.float()
                coords_i = coords_i.float()
            (corr,) = alt_cuda_corr.forward(fmap1_i, fmap2_i, coords_i, r)
            if coords.dtype == torch.float16:
                corr = corr.half()
            corr_list.append(corr.squeeze(1))

        corr = torch.stack(corr_list, dim=1)
        corr = corr.reshape(B, -1, H, W)
        return corr / torch.sqrt(torch.tensor(dim).float())


class CorrBlock1D:
    def __init__(self, corr1, corr2, num_levels=4, radius=4):
        self.num_levels = num_levels
        self.radius = radius
        self.corr_pyramid1 = []
        self.corr_pyramid2 = []

        corr1 = corr1.permute(0, 3, 4, 1, 2)
        corr2 = corr2.permute(0, 3, 4, 1, 2)
        batch, h1, w1, dim, w2 = corr1.shape
        batch, h1, w1, dim, h2 = corr2.shape
        assert corr1.shape[:-1] == corr2.shape[:-1]
        assert h1 == h2 and w1 == w2

        # self.coords = coords_grid(batch, h2, w2).to(corr1.device)

        corr1 = corr1.reshape(batch * h1 * w1, dim, 1, w2)
        corr2 = corr2.reshape(batch * h1 * w1, dim, 1, h2)

        self.corr_pyramid1.append(corr1)
        self.corr_pyramid2.append(corr2)
        for i in range(self.num_levels):
            corr1 = F.avg_pool2d(corr1, [1, 2], stride=[1, 2])
            self.corr_pyramid1.append(corr1)
            corr2 = F.avg_pool2d(corr2, [1, 2], stride=[1, 2])
            self.corr_pyramid2.append(corr2)
            # print(corr1.shape, corr1.mean().item(), corr2.shape, corr2.mean().item())

    def bilinear_sampler(self, img, coords, mode="bilinear", mask=False):
        """Wrapper for grid_sample, uses pixel coordinates"""
        H, W = img.shape[-2:]
        xgrid, ygrid = coords.split([1, 1], dim=-1)
        xgrid = 2 * xgrid / (W - 1) - 1
        assert torch.unique(ygrid).numel() == 1 and H == 1  # This is a stereo problem

        grid = torch.cat([xgrid, ygrid], dim=-1)
        img = F.grid_sample(img, grid, align_corners=True)

        if mask:
            mask = (xgrid > -1) & (ygrid > -1) & (xgrid < 1) & (ygrid < 1)
            return img, mask.float()

        return img

    def __call__(self, coords):
        coords_org = coords.clone()
        coords = coords_org[:, :1, :, :]
        coords = coords.permute(0, 2, 3, 1)
        r = self.radius
        batch, h1, w1, _ = coords.shape

        out_pyramid = []
        for i in range(self.num_levels):
            corr = self.corr_pyramid1[i]
            dx = torch.linspace(-r, r, 2 * r + 1)
            dx = dx.view(1, 1, 2 * r + 1, 1).to(coords.device)
            x0 = dx + coords.reshape(batch * h1 * w1, 1, 1, 1) / 2**i
            y0 = torch.zeros_like(x0)

            coords_lvl = torch.cat([x0, y0], dim=-1)
            coords_lvl = torch.clamp(coords_lvl, -1, 1)
            # print("corri:", corr.shape, corr.mean().item(), coords_lvl.shape, coords_lvl.mean().item())
            corr = self.bilinear_sampler(corr, coords_lvl)
            # print("corri:", corr.shape, corr.mean().item())
            corr = corr.view(batch, h1, w1, -1)
            # print("corri:", corr.shape, corr.mean().item())
            out_pyramid.append(corr)

        out = torch.cat(out_pyramid, dim=-1)
        out1 = out.permute(0, 3, 1, 2).contiguous().float()

        coords = coords_org[:, 1:, :, :]
        coords = coords.permute(0, 2, 3, 1)
        r = self.radius
        batch, h1, w1, _ = coords.shape

        out_pyramid = []
        for i in range(self.num_levels):
            corr = self.corr_pyramid2[i]
            dx = torch.linspace(-r, r, 2 * r + 1)
            dx = dx.view(1, 1, 2 * r + 1, 1).to(coords.device)
            x0 = dx + coords.reshape(batch * h1 * w1, 1, 1, 1) / 2**i
            y0 = torch.zeros_like(x0)

            coords_lvl = torch.cat([x0, y0], dim=-1)
            corr = self.bilinear_sampler(corr, coords_lvl)
            corr = corr.view(batch, h1, w1, -1)
            out_pyramid.append(corr)

        out = torch.cat(out_pyramid, dim=-1)
        out2 = out.permute(0, 3, 1, 2).contiguous().float()
        return out1, out2



from .constants import *

# from .auto_augment import RandAugment, AutoAugment, rand_augment_ops, auto_augment_policy,\
#     rand_augment_transform, auto_augment_transform
# from .config import resolve_data_config
# from .dataset import ImageDataset, IterableImageDataset, AugMixDataset
# from .dataset_factory import create_dataset
# from .loader import create_loader
# from .mixup import Mixup, FastCollateMixup
# from .parsers import create_parser
# from .real_labels import RealLabelsImagenet
# from .transforms import *
# from .transforms_factory import create_transform



from typing import Any, Dict, List, Optional

try:
    from spatial_correlation_sampler import SpatialCorrelationSampler
except ModuleNotFoundError:
    from ptlflow.utils.correlation import (
        IterSpatialCorrelationSampler as SpatialCorrelationSampler,
    )
import torch
import torch.nn as nn
import torch.nn.functional as F

from ptlflow.utils.registry import register_model
from .warp import WarpingLayer
from ..base_model.base_model import BaseModel


class FeatureExtractor(nn.Module):
    def __init__(self):
        super(FeatureExtractor, self).__init__()

        leaky_relu = nn.LeakyReLU(0.1, inplace=True)

        self.convs = nn.ModuleList(
            [
                nn.Sequential(nn.Conv2d(3, 32, 7, 1, 3), leaky_relu),
                nn.Sequential(
                    nn.Conv2d(32, 32, 3, 2, 1),
                    leaky_relu,
                    nn.Conv2d(32, 32, 3, 1, 1),
                    leaky_relu,
                    nn.Conv2d(32, 32, 3, 1, 1),
                    leaky_relu,
                ),
                nn.Sequential(
                    nn.Conv2d(32, 64, 3, 2, 1),
                    leaky_relu,
                    nn.Conv2d(64, 64, 3, 1, 1),
                    leaky_relu,
                ),
                nn.Sequential(
                    nn.Conv2d(64, 96, 3, 2, 1),
                    leaky_relu,
                    nn.Conv2d(96, 96, 3, 1, 1),
                    leaky_relu,
                ),
                nn.Sequential(nn.Conv2d(96, 128, 3, 2, 1), leaky_relu),
                nn.Sequential(nn.Conv2d(128, 192, 3, 2, 1), leaky_relu),
            ]
        )

    def forward(self, images: torch.Tensor) -> List[torch.Tensor]:
        features = []

        x = images.view(-1, *images.shape[2:])
        for i, conv in enumerate(self.convs):
            x = conv(x)
            if i > 1:
                features.append(x.view(*images.shape[:2], *x.shape[1:]))

        return features[::-1]


class Matching(nn.Module):
    def __init__(self, level: int, num_levels: int = 4, div_flow: float = 20.0) -> None:
        super(Matching, self).__init__()

        corr_stride = [1, 1, 1, 2][level]
        flow_kernel_size = [3, 3, 5, 5][level]
        self.mult = [div_flow / 2 ** (num_levels - i + 1) for i in range(num_levels)][
            level
        ]

        self.leaky_relu = nn.LeakyReLU(0.1, inplace=True)

        if level == 0:
            self.up_flow = None
        else:
            self.up_flow = nn.ConvTranspose2d(2, 2, 4, 2, 1, bias=False, groups=2)

        if level < 3:
            self.up_corr = None
        else:
            self.up_corr = nn.ConvTranspose2d(49, 49, 4, 2, 1, bias=False, groups=49)

        self.flow_net = nn.Sequential(
            nn.Conv2d(49, 128, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(128, 128, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(128, 96, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(96, 64, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(64, 32, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(32, 2, flow_kernel_size, 1, flow_kernel_size // 2),
        )

        self.corr = SpatialCorrelationSampler(
            kernel_size=1,
            patch_size=7,
            padding=0,
            stride=corr_stride,
            dilation_patch=corr_stride,
        )

        self.warp = WarpingLayer()

    def forward(
        self, feats: torch.Tensor, flow: Optional[torch.Tensor]
    ) -> torch.Tensor:
        warped_feat2 = feats[:, 1]
        if flow is not None:
            flow = self.up_flow(flow)
            warped_feat2 = self.warp(
                feats[:, 1], flow, feats.shape[-2], feats.shape[-1], 1.0 / self.mult
            )

        corr = self.leaky_relu(self.corr(feats[:, 0], warped_feat2))
        corr = corr.view(corr.shape[0], -1, corr.shape[3], corr.shape[4])
        corr = corr / feats.shape[2]
        if self.up_corr is not None:
            corr = self.up_corr(corr)
        new_flow = self.flow_net(corr)
        if flow is not None:
            new_flow = flow + new_flow
        return new_flow


class SubPixel(nn.Module):
    def __init__(self, level: int, num_levels: int = 4, div_flow: float = 20.0) -> None:
        super(SubPixel, self).__init__()

        inputs_dims = [386, 258, 194, 130][level]
        flow_kernel_size = [3, 3, 5, 5][level]
        self.mult = [div_flow / 2 ** (num_levels - i + 1) for i in range(num_levels)][
            level
        ]

        self.leaky_relu = nn.LeakyReLU(0.1, inplace=True)

        self.feat_net = nn.Sequential(
            nn.Conv2d(inputs_dims, 128, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(128, 128, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(128, 96, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(96, 64, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(64, 32, 3, 1, 1),
            self.leaky_relu,
        )

        self.flow_net = nn.Conv2d(32, 2, flow_kernel_size, 1, flow_kernel_size // 2)

        self.warp = WarpingLayer()

    def forward(self, feats: torch.Tensor, flow: torch.Tensor) -> torch.Tensor:
        feat_warped = self.warp(
            feats[:, 1], flow, feats.shape[-2], feats.shape[-1], 1.0 / self.mult
        )
        x = torch.cat([feats[:, 0], feat_warped, flow], dim=1)
        x = self.feat_net(x)
        new_flow = self.flow_net(x)
        new_flow = flow + new_flow
        return new_flow, x


class Regularization(nn.Module):
    def __init__(self, level: int, num_levels: int = 4, div_flow: float = 20.0) -> None:
        super(Regularization, self).__init__()

        inputs_dims = [195, 131, 99, 67][level]
        flow_kernel_size = [3, 3, 5, 5][level]
        self.mult = [div_flow / 2 ** (num_levels - i + 1) for i in range(num_levels)][
            level
        ]

        self.leaky_relu = nn.LeakyReLU(0.1, inplace=True)

        if level < 2:
            self.feat_conv = nn.Sequential()
        else:
            self.feat_conv = nn.Sequential(
                nn.Conv2d(inputs_dims - 3, 128, 1, 1, 0), self.leaky_relu
            )
            inputs_dims = 131

        self.feat_net = nn.Sequential(
            nn.Conv2d(inputs_dims, 128, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(128, 128, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(128, 64, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(64, 64, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(64, 32, 3, 1, 1),
            self.leaky_relu,
            nn.Conv2d(32, 32, 3, 1, 1),
            self.leaky_relu,
        )

        if level < 2:
            self.dist = nn.Conv2d(32, flow_kernel_size**2, 3, 1, 1)
        else:
            self.dist = nn.Sequential(
                nn.Conv2d(
                    32,
                    flow_kernel_size**2,
                    (flow_kernel_size, 1),
                    1,
                    (flow_kernel_size // 2, 0),
                ),
                nn.Conv2d(
                    flow_kernel_size**2,
                    flow_kernel_size**2,
                    (1, flow_kernel_size),
                    1,
                    (0, flow_kernel_size // 2),
                ),
            )

        self.unfold = nn.Unfold(flow_kernel_size, padding=flow_kernel_size // 2)

        self.warp = WarpingLayer()

    def forward(
        self, images: torch.Tensor, feats: torch.Tensor, flow: torch.Tensor
    ) -> torch.Tensor:
        img2_warped = self.warp(
            images[:, 1], flow, images.shape[-2], images.shape[-1], 1.0 / self.mult
        )
        img_diff_norm = torch.norm(
            images[:, 0] - img2_warped[:, 1], p=2, dim=1, keepdim=True
        )

        flow_mean = flow.view(*flow.shape[:2], -1).mean(dim=-1)[..., None, None]
        flow_nomean = flow - flow_mean
        feat = self.feat_conv(feats[:, 0])
        x = torch.cat([img_diff_norm, flow_nomean, feat], dim=1)
        x = self.feat_net(x)
        dist = self.dist(x)
        dist = dist.square().neg()
        dist = (dist - dist.max(dim=1, keepdim=True)[0]).exp()
        div = dist.sum(dim=1, keepdim=True)

        reshaped_flow_x = self.unfold(flow[:, :1])
        reshaped_flow_x = reshaped_flow_x.view(
            *reshaped_flow_x.shape[:2], *flow.shape[2:4]
        )
        flow_smooth_x = (reshaped_flow_x * dist).sum(dim=1, keepdim=True) / div

        reshaped_flow_y = self.unfold(flow[:, 1:2])
        reshaped_flow_y = reshaped_flow_y.view(
            *reshaped_flow_y.shape[:2], *flow.shape[2:4]
        )
        flow_smooth_y = (reshaped_flow_y * dist).sum(dim=1, keepdim=True) / div

        flow = torch.cat([flow_smooth_x, flow_smooth_y], dim=1)

        return flow, x


class PseudoSubpixel(nn.Module):
    def __init__(self) -> None:
        super(PseudoSubpixel, self).__init__()

        self.up_flow = nn.ConvTranspose2d(2, 2, 4, 2, 1, bias=False, groups=2)

        self.flow_net = nn.Sequential(
            nn.ConvTranspose2d(32, 32, 4, 2, 1), nn.Conv2d(32, 2, 7, 1, 3)
        )

    def forward(self, sub_feat: torch.Tensor, flow: torch.Tensor) -> torch.Tensor:
        return self.up_flow(flow) + self.flow_net(sub_feat)


class PseudoRegularization(nn.Module):
    def __init__(self) -> None:
        super(PseudoRegularization, self).__init__()

        self.feat_net = nn.Sequential(
            nn.ConvTranspose2d(32, 32, 4, 2, 1),
            nn.Conv2d(32, 49, (7, 1), 1, (3, 0)),
            nn.Conv2d(49, 49, (1, 7), 1, (0, 3)),
        )

        self.unfold = nn.Unfold(7, padding=3)

    def forward(self, reg_feat: torch.Tensor, flow: torch.Tensor) -> torch.Tensor:
        dist = self.feat_net(reg_feat)
        dist = dist.square().neg()
        dist = (dist - dist.max(dim=1, keepdim=True)[0]).exp()
        div = dist.sum(dim=1, keepdim=True)

        reshaped_flow_x = self.unfold(flow[:, :1])
        reshaped_flow_x = reshaped_flow_x.view(
            *reshaped_flow_x.shape[:2], *flow.shape[2:4]
        )
        flow_smooth_x = (reshaped_flow_x * dist).sum(dim=1, keepdim=True) / div

        reshaped_flow_y = self.unfold(flow[:, 1:2])
        reshaped_flow_y = reshaped_flow_y.view(
            *reshaped_flow_y.shape[:2], *flow.shape[2:4]
        )
        flow_smooth_y = (reshaped_flow_y * dist).sum(dim=1, keepdim=True) / div

        flow = torch.cat([flow_smooth_x, flow_smooth_y], dim=1)

        return flow


class LiteFlowNet2(BaseModel):
    pretrained_checkpoints = {
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/liteflownet2-sintel-1e1eb282.ckpt"
    }

    def __init__(
        self,
        div_flow: float = 20.0,
        use_pseudo_regularization: bool = False,
        **kwargs,
    ):
        super(LiteFlowNet2, self).__init__(
            loss_fn=None,
            output_stride=32,
            **kwargs,
        )

        self.div_flow = div_flow
        self.use_pseudo_regularization = use_pseudo_regularization

        self.num_levels = 4

        self.feature_net = FeatureExtractor()
        self.matching_nets = nn.ModuleList(
            [
                Matching(i, self.num_levels, self.div_flow)
                for i in range(self.num_levels)
            ]
        )
        self.subpixel_nets = nn.ModuleList(
            [
                SubPixel(i, self.num_levels, self.div_flow)
                for i in range(self.num_levels)
            ]
        )
        self.regularization_nets = nn.ModuleList(
            [
                Regularization(i, self.num_levels, self.div_flow)
                for i in range(self.num_levels)
            ]
        )

        if self.use_pseudo_regularization:
            self.pseudo_subpixel = PseudoSubpixel()
            self.pseudo_regularization = PseudoRegularization()
            self.up_flow = nn.ConvTranspose2d(2, 2, 4, 2, 1, bias=False, groups=2)
        else:
            self.up_flow = nn.ConvTranspose2d(2, 2, 8, 4, 2, bias=False, groups=2)

    def forward(self, inputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        # The original implementation uses different BGR means for im1 and im2
        # Here, we use the same for both to make it simpler
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=[-0.454253, -0.434631, -0.411618],
            bgr_mult=1.0,
            bgr_to_rgb=True,
            resize_mode="interpolation",
            interpolation_mode="bilinear",
            interpolation_align_corners=False,
        )

        feats_pyr = self.feature_net(images)
        images_pyr = self._create_images_pyr(images, feats_pyr)

        flow_preds = []
        flow = None

        for i in range(self.num_levels):
            flow = self.matching_nets[i](feats_pyr[i], flow)
            flow, sub_feat = self.subpixel_nets[i](feats_pyr[i], flow)
            flow, reg_feat = self.regularization_nets[i](
                images_pyr[i], feats_pyr[i], flow
            )
            flow_preds.append(flow)

        if self.use_pseudo_regularization:
            flow = self.pseudo_subpixel(sub_feat, flow)
            flow = self.pseudo_regularization(reg_feat, flow)
            flow = self.up_flow(flow)
        else:
            flow = self.up_flow(flow)
        flow = flow * self.div_flow
        flow = self.postprocess_predictions(flow, image_resizer, is_flow=True)

        outputs = {}
        if self.training:
            outputs["flow_preds"] = flow_preds
            outputs["flows"] = flow[:, None]
        else:
            outputs["flows"] = flow[:, None]
        return outputs

    def _create_images_pyr(
        self, images: torch.Tensor, feats_pyr: List[torch.Tensor]
    ) -> List[torch.Tensor]:
        batch_size = images.shape[0]
        images = images.view(-1, *images.shape[2:]).detach()
        images_pyr = [
            F.interpolate(
                images,
                size=feats_pyr[i].shape[-2:],
                mode="bilinear",
                align_corners=False,
            )
            for i in range(len(feats_pyr))
        ]
        images_pyr = [im.view(batch_size, -1, *im.shape[1:]) for im in images_pyr]
        return images_pyr


class LiteFlowNet2PseudoReg(LiteFlowNet2):
    pretrained_checkpoints = {
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/liteflownet2-kitti-da069fca.ckpt"
    }

    def __init__(
        self,
        div_flow: float = 20,
        use_pseudo_regularization: bool = True,
        **kwargs,
    ):
        super().__init__(
            div_flow,
            use_pseudo_regularization,
            **kwargs,
        )


@register_model
class liteflownet2(LiteFlowNet2):
    pass


@register_model
class liteflownet2_pseudoreg(LiteFlowNet2PseudoReg):
    pass



import torch
import torch.nn as nn
from .layer import BasicBlock, conv1x1


class ResNetFPN(nn.Module):
    """
    ResNet18, output resolution is 1/8.
    Each block has 2 layers.
    """

    def __init__(
        self,
        block_dims,
        initial_dim,
        pretrain,
        input_dim=3,
        output_dim=256,
        ratio=1.0,
        norm_layer=nn.BatchNorm2d,
        init_weight=False,
    ):
        super().__init__()
        # Config
        block = BasicBlock
        block_dims = block_dims
        initial_dim = initial_dim
        self.init_weight = init_weight
        self.input_dim = input_dim
        # Class Variable
        self.in_planes = initial_dim
        for i in range(len(block_dims)):
            block_dims[i] = int(block_dims[i] * ratio)
        # Networks
        self.conv1 = nn.Conv2d(
            input_dim, initial_dim, kernel_size=7, stride=2, padding=3
        )
        self.bn1 = norm_layer(initial_dim)
        self.relu = nn.ReLU(inplace=True)
        if pretrain == "resnet34":
            n_block = [3, 4, 6]
        elif pretrain == "resnet18":
            n_block = [2, 2, 2]
        else:
            raise NotImplementedError
        self.layer1 = self._make_layer(
            block, block_dims[0], stride=1, norm_layer=norm_layer, num=n_block[0]
        )  # 1/2
        self.layer2 = self._make_layer(
            block, block_dims[1], stride=2, norm_layer=norm_layer, num=n_block[1]
        )  # 1/4
        self.layer3 = self._make_layer(
            block, block_dims[2], stride=2, norm_layer=norm_layer, num=n_block[2]
        )  # 1/8
        self.final_conv = conv1x1(block_dims[2], output_dim)
        self._init_weights(pretrain)

    def _init_weights(self, pretrain):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, (nn.BatchNorm2d, nn.InstanceNorm2d, nn.GroupNorm)):
                if m.weight is not None:
                    nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

        if self.init_weight:
            from torchvision.models import (
                resnet18,
                ResNet18_Weights,
                resnet34,
                ResNet34_Weights,
            )

            if pretrain == "resnet18":
                pretrained_dict = resnet18(
                    weights=ResNet18_Weights.IMAGENET1K_V1
                ).state_dict()
            else:
                pretrained_dict = resnet34(
                    weights=ResNet34_Weights.IMAGENET1K_V1
                ).state_dict()
            model_dict = self.state_dict()
            pretrained_dict = {
                k: v for k, v in pretrained_dict.items() if k in model_dict
            }
            if self.input_dim == 6:
                for k, v in pretrained_dict.items():
                    if k == "conv1.weight":
                        pretrained_dict[k] = torch.cat((v, v), dim=1)
            model_dict.update(pretrained_dict)
            self.load_state_dict(model_dict, strict=False)

    def _make_layer(self, block, dim, stride=1, norm_layer=nn.BatchNorm2d, num=2):
        layers = []
        layers.append(block(self.in_planes, dim, stride=stride, norm_layer=norm_layer))
        for i in range(num - 1):
            layers.append(block(dim, dim, stride=1, norm_layer=norm_layer))
        self.in_planes = dim
        return nn.Sequential(*layers)

    def forward(self, x):
        # ResNet Backbone
        x = self.relu(self.bn1(self.conv1(x)))
        for i in range(len(self.layer1)):
            x = self.layer1[i](x)
        for i in range(len(self.layer2)):
            x = self.layer2[i](x)
        for i in range(len(self.layer3)):
            x = self.layer3[i](x)
        # Output
        output = self.final_conv(x)
        return output



"""
Portions of this code copyright 2017, Clement Pinard
"""

from einops import rearrange
import torch
import torch.nn as nn
from torch.nn import init

from ptlflow.utils.registry import register_model, trainable
from .submodules import *
from .flownet_base import FlowNetBase


class FlowNetSD(FlowNetBase):
    pretrained_checkpoints = {
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/flownetsd-things-c5f3124e.ckpt"
    }

    def __init__(
        self,
        div_flow: float = 20.0,
        input_channels: int = 6,
        batch_norm: bool = False,
        loss_start_scale: int = 4,
        loss_num_scales: int = 5,
        loss_base_weight: float = 0.32,
        loss_norm: str = "L2",
        **kwargs,
    ):
        super(FlowNetSD, self).__init__(
            div_flow=div_flow,
            input_channels=input_channels,
            batch_norm=batch_norm,
            loss_start_scale=loss_start_scale,
            loss_num_scales=loss_num_scales,
            loss_base_weight=loss_base_weight,
            loss_norm=loss_norm,
            **kwargs,
        )

        self.conv0 = conv(self.batch_norm, 6, 64)
        self.conv1 = conv(self.batch_norm, 64, 64, stride=2)
        self.conv1_1 = conv(self.batch_norm, 64, 128)
        self.conv2 = conv(self.batch_norm, 128, 128, stride=2)
        self.conv2_1 = conv(self.batch_norm, 128, 128)
        self.conv3 = conv(self.batch_norm, 128, 256, stride=2)
        self.conv3_1 = conv(self.batch_norm, 256, 256)
        self.conv4 = conv(self.batch_norm, 256, 512, stride=2)
        self.conv4_1 = conv(self.batch_norm, 512, 512)
        self.conv5 = conv(self.batch_norm, 512, 512, stride=2)
        self.conv5_1 = conv(self.batch_norm, 512, 512)
        self.conv6 = conv(self.batch_norm, 512, 1024, stride=2)
        self.conv6_1 = conv(self.batch_norm, 1024, 1024)

        self.deconv5 = deconv(1024, 512)
        self.deconv4 = deconv(1026, 256)
        self.deconv3 = deconv(770, 128)
        self.deconv2 = deconv(386, 64)

        self.inter_conv5 = i_conv(self.batch_norm, 1026, 512)
        self.inter_conv4 = i_conv(self.batch_norm, 770, 256)
        self.inter_conv3 = i_conv(self.batch_norm, 386, 128)
        self.inter_conv2 = i_conv(self.batch_norm, 194, 64)

        self.predict_flow6 = predict_flow(1024)
        self.predict_flow5 = predict_flow(512)
        self.predict_flow4 = predict_flow(256)
        self.predict_flow3 = predict_flow(128)
        self.predict_flow2 = predict_flow(64)

        self.upsampled_flow6_to_5 = nn.ConvTranspose2d(2, 2, 4, 2, 1)
        self.upsampled_flow5_to_4 = nn.ConvTranspose2d(2, 2, 4, 2, 1)
        self.upsampled_flow4_to_3 = nn.ConvTranspose2d(2, 2, 4, 2, 1)
        self.upsampled_flow3_to_2 = nn.ConvTranspose2d(2, 2, 4, 2, 1)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                if m.bias is not None:
                    init.uniform_(m.bias)
                init.xavier_uniform_(m.weight)

            if isinstance(m, nn.ConvTranspose2d):
                if m.bias is not None:
                    init.uniform_(m.bias)
                init.xavier_uniform_(m.weight)
                # init_deconv_bilinear(m.weight)
        self.upsample1 = nn.Upsample(
            scale_factor=4, mode="bilinear", align_corners=False
        )

    def forward(self, inputs, skip_preprocess=False):
        images = inputs["images"]
        if skip_preprocess:
            image_resizer = None
        else:
            bgr_mean = rearrange(images, "b n c h w -> b n c (h w)").mean(-1)
            bgr_mean = bgr_mean[..., None, None]
            images, image_resizer = self.preprocess_images(
                images,
                bgr_add=-bgr_mean,
                bgr_mult=1.0,
                bgr_to_rgb=True,
                resize_mode="interpolation",
                interpolation_mode="bilinear",
                interpolation_align_corners=True,
            )
        x = images

        x = x.view(x.shape[0], x.shape[1] * x.shape[2], x.shape[3], x.shape[4])
        out_conv0 = self.conv0(x)
        out_conv1 = self.conv1_1(self.conv1(out_conv0))
        out_conv2 = self.conv2_1(self.conv2(out_conv1))

        out_conv3 = self.conv3_1(self.conv3(out_conv2))
        out_conv4 = self.conv4_1(self.conv4(out_conv3))
        out_conv5 = self.conv5_1(self.conv5(out_conv4))
        out_conv6 = self.conv6_1(self.conv6(out_conv5))

        flow6 = self.predict_flow6(out_conv6)
        flow6_up = self.upsampled_flow6_to_5(flow6)
        out_deconv5 = self.deconv5(out_conv6)

        concat5 = torch.cat((out_conv5, out_deconv5, flow6_up), 1)
        out_interconv5 = self.inter_conv5(concat5)
        flow5 = self.predict_flow5(out_interconv5)

        flow5_up = self.upsampled_flow5_to_4(flow5)
        out_deconv4 = self.deconv4(concat5)

        concat4 = torch.cat((out_conv4, out_deconv4, flow5_up), 1)
        out_interconv4 = self.inter_conv4(concat4)
        flow4 = self.predict_flow4(out_interconv4)
        flow4_up = self.upsampled_flow4_to_3(flow4)
        out_deconv3 = self.deconv3(concat4)

        concat3 = torch.cat((out_conv3, out_deconv3, flow4_up), 1)
        out_interconv3 = self.inter_conv3(concat3)
        flow3 = self.predict_flow3(out_interconv3)
        flow3_up = self.upsampled_flow3_to_2(flow3)
        out_deconv2 = self.deconv2(concat3)

        concat2 = torch.cat((out_conv2, out_deconv2, flow3_up), 1)
        out_interconv2 = self.inter_conv2(concat2)
        flow2 = self.predict_flow2(out_interconv2)

        # Results are noticeably better if dividing here, rather than multiplying
        # Maybe it was trained with the inverse divisor?
        out_flow = self.upsample1(flow2) / self.div_flow
        if image_resizer is not None:
            out_flow = self.postprocess_predictions(
                out_flow, image_resizer, is_flow=True
            )

        outputs = {}

        if self.training:
            outputs["flow_preds"] = [
                flow2,
                flow3,
                flow4,
                flow5,
                flow6,
            ]
            outputs["flows"] = out_flow[:, None]
        else:
            outputs["flows"] = out_flow[:, None]

        return outputs


@register_model
@trainable
class flownetsd(FlowNetSD):
    pass



import torch
import torch.nn as nn
import torch.nn.functional as F


class GELU(nn.Module):
    """Applies the Gaussian Error Linear Units function (w/ dummy inplace arg)"""

    def __init__(self, inplace: bool = False):
        super(GELU, self).__init__()

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return F.gelu(input)



import torch
import torch.nn.functional as F

from .geometry import coords_grid, generate_window_grid, normalize_coords


def global_correlation_softmax(
    feature0,
    feature1,
    pred_bidir_flow=False,
):
    # global correlation
    b, c, h, w = feature0.shape
    feature0 = feature0.view(b, c, -1).permute(0, 2, 1)  # [B, H*W, C]
    feature1 = feature1.view(b, c, -1)  # [B, C, H*W]

    correlation = torch.matmul(feature0, feature1).view(b, h, w, h, w) / (
        c**0.5
    )  # [B, H, W, H, W]

    # flow from softmax
    init_grid = coords_grid(
        b, h, w, dtype=correlation.dtype, device=correlation.device
    )  # [B, 2, H, W]
    grid = init_grid.view(b, 2, -1).permute(0, 2, 1)  # [B, H*W, 2]

    correlation = correlation.view(b, h * w, h * w)  # [B, H*W, H*W]

    if pred_bidir_flow:
        correlation = torch.cat(
            (correlation, correlation.permute(0, 2, 1)), dim=0
        )  # [2*B, H*W, H*W]
        init_grid = init_grid.repeat(2, 1, 1, 1)  # [2*B, 2, H, W]
        grid = grid.repeat(2, 1, 1)  # [2*B, H*W, 2]
        b = b * 2

    prob = F.softmax(correlation, dim=-1)  # [B, H*W, H*W]

    correspondence = (
        torch.matmul(prob, grid).view(b, h, w, 2).permute(0, 3, 1, 2)
    )  # [B, 2, H, W]

    # when predicting bidirectional flow, flow is the concatenation of forward flow and backward flow
    flow = correspondence - init_grid

    return flow, prob


def local_correlation_softmax(
    feature0,
    feature1,
    local_radius,
    padding_mode="zeros",
):
    b, c, h, w = feature0.size()
    coords_init = coords_grid(
        b, h, w, dtype=feature0.dtype, device=feature0.device
    )  # [B, 2, H, W]
    coords = coords_init.view(b, 2, -1).permute(0, 2, 1)  # [B, H*W, 2]

    local_h = 2 * local_radius + 1
    local_w = 2 * local_radius + 1

    window_grid = generate_window_grid(
        -local_radius,
        local_radius,
        -local_radius,
        local_radius,
        local_h,
        local_w,
        dtype=feature0.dtype,
        device=feature0.device,
    )  # [2R+1, 2R+1, 2]
    window_grid = window_grid.reshape(-1, 2).repeat(b, 1, 1, 1)  # [B, 1, (2R+1)^2, 2]
    sample_coords = coords.unsqueeze(-2) + window_grid  # [B, H*W, (2R+1)^2, 2]

    sample_coords_softmax = sample_coords

    # exclude coords that are out of image space
    valid_x = (sample_coords[:, :, :, 0] >= 0) & (
        sample_coords[:, :, :, 0] < w
    )  # [B, H*W, (2R+1)^2]
    valid_y = (sample_coords[:, :, :, 1] >= 0) & (
        sample_coords[:, :, :, 1] < h
    )  # [B, H*W, (2R+1)^2]

    valid = (
        valid_x & valid_y
    )  # [B, H*W, (2R+1)^2], used to mask out invalid values when softmax

    # normalize coordinates to [-1, 1]
    sample_coords_norm = normalize_coords(sample_coords, h, w)  # [-1, 1]
    window_feature = F.grid_sample(
        feature1, sample_coords_norm, padding_mode=padding_mode, align_corners=True
    ).permute(
        0, 2, 1, 3
    )  # [B, H*W, C, (2R+1)^2]
    feature0_view = feature0.permute(0, 2, 3, 1).view(b, h * w, 1, c)  # [B, H*W, 1, C]

    corr = torch.matmul(feature0_view, window_feature).view(b, h * w, -1) / (
        c**0.5
    )  # [B, H*W, (2R+1)^2]

    # mask invalid locations
    if feature0.dtype == torch.float16:
        val = -1e4
    else:
        val = -1e9
    corr[~valid] = val

    prob = F.softmax(corr, -1)  # [B, H*W, (2R+1)^2]

    correspondence = (
        torch.matmul(prob.unsqueeze(-2), sample_coords_softmax)
        .squeeze(-2)
        .view(b, h, w, 2)
        .permute(0, 3, 1, 2)
    )  # [B, 2, H, W]

    flow = correspondence - coords_init
    match_prob = prob

    return flow, match_prob


def local_correlation_with_flow(
    feature0,
    feature1,
    flow,
    local_radius,
    padding_mode="zeros",
    dilation=1,
):
    b, c, h, w = feature0.size()
    coords_init = coords_grid(
        b, h, w, dtype=feature0.dtype, device=feature0.device
    )  # [B, 2, H, W]
    coords = coords_init.view(b, 2, -1).permute(0, 2, 1)  # [B, H*W, 2]

    local_h = 2 * local_radius + 1
    local_w = 2 * local_radius + 1

    window_grid = generate_window_grid(
        -local_radius,
        local_radius,
        -local_radius,
        local_radius,
        local_h,
        local_w,
        dtype=feature0.dtype,
        device=feature0.device,
    )  # [2R+1, 2R+1, 2]
    window_grid = window_grid.reshape(-1, 2).repeat(b, 1, 1, 1)  # [B, 1, (2R+1)^2, 2]
    sample_coords = (
        coords.unsqueeze(-2) + window_grid * dilation
    )  # [B, H*W, (2R+1)^2, 2]

    # flow can be zero when using features after transformer
    if not isinstance(flow, float):
        sample_coords = sample_coords + flow.view(b, 2, -1).permute(0, 2, 1).unsqueeze(
            -2
        )  # [B, H*W, (2R+1)^2, 2]
    else:
        assert flow == 0.0

    # normalize coordinates to [-1, 1]
    sample_coords_norm = normalize_coords(sample_coords, h, w)  # [-1, 1]
    window_feature = F.grid_sample(
        feature1, sample_coords_norm, padding_mode=padding_mode, align_corners=True
    ).permute(
        0, 2, 1, 3
    )  # [B, H*W, C, (2R+1)^2]
    feature0_view = feature0.permute(0, 2, 3, 1).view(b, h * w, 1, c)  # [B, H*W, 1, C]

    corr = torch.matmul(feature0_view, window_feature).view(b, h * w, -1) / (
        c**0.5
    )  # [B, H*W, (2R+1)^2]

    corr = (
        corr.view(b, h, w, -1).permute(0, 3, 1, 2).contiguous()
    )  # [B, (2R+1)^2, H, W]

    return corr


def warp_with_pose_depth_candidates(
    feature1,
    intrinsics,
    pose,
    depth,
    clamp_min_depth=1e-3,
):
    """
    feature1: [B, C, H, W]
    intrinsics: [B, 3, 3]
    pose: [B, 4, 4]
    depth: [B, D, H, W]
    """

    assert intrinsics.size(1) == intrinsics.size(2) == 3
    assert pose.size(1) == pose.size(2) == 4
    assert depth.dim() == 4

    b, d, h, w = depth.size()
    c = feature1.size(1)

    with torch.no_grad():
        # pixel coordinates
        grid = coords_grid(
            b, h, w, homogeneous=True, dtype=depth.dtype, device=depth.device
        )  # [B, 3, H, W]
        # back project to 3D and transform viewpoint
        points = torch.inverse(intrinsics).bmm(grid.view(b, 3, -1))  # [B, 3, H*W]
        points = torch.bmm(pose[:, :3, :3], points).unsqueeze(2).repeat(
            1, 1, d, 1
        ) * depth.view(
            b, 1, d, h * w
        )  # [B, 3, D, H*W]
        points = points + pose[:, :3, -1:].unsqueeze(-1)  # [B, 3, D, H*W]
        # reproject to 2D image plane
        points = torch.bmm(intrinsics, points.view(b, 3, -1)).view(
            b, 3, d, h * w
        )  # [B, 3, D, H*W]
        pixel_coords = points[:, :2] / points[:, -1:].clamp(
            min=clamp_min_depth
        )  # [B, 2, D, H*W]

        # normalize to [-1, 1]
        x_grid = 2 * pixel_coords[:, 0] / (w - 1) - 1
        y_grid = 2 * pixel_coords[:, 1] / (h - 1) - 1

        grid = torch.stack([x_grid, y_grid], dim=-1)  # [B, D, H*W, 2]

    # sample features
    warped_feature = F.grid_sample(
        feature1,
        grid.view(b, d * h, w, 2),
        mode="bilinear",
        padding_mode="zeros",
        align_corners=True,
    ).view(
        b, c, d, h, w
    )  # [B, C, D, H, W]

    return warped_feature



import loguru
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import einsum

from einops.layers.torch import Rearrange
from einops import rearrange

from ..encoders import twins_svt_large, convnext_large
from .twins import PosConv
from .encoder import MemoryEncoder
from .decoder import MemoryDecoder
from .cnn import BasicEncoder


class FlowFormer(nn.Module):
    def __init__(self, cfg):
        super(FlowFormer, self).__init__()

        H1, W1, H2, W2 = cfg.pic_size
        H_offset = (H1 - H2) // 2
        W_offset = (W1 - W2) // 2
        cfg.H_offset = H_offset
        cfg.W_offset = W_offset

        self.cfg = cfg

        self.memory_encoder = MemoryEncoder(cfg)
        self.memory_decoder = MemoryDecoder(cfg)
        if cfg.cnet == "twins":
            self.context_encoder = twins_svt_large(
                pretrained=self.cfg.pretrain, del_layers=cfg.del_layers
            )
        elif "jh" in cfg.cnet:
            self.context_encoder = twins_svt_large_jihao(
                pretrained=self.cfg.pretrain,
                del_layers=cfg.del_layers,
                version=cfg.cnet,
            )
        elif cfg.cnet == "basicencoder":
            self.context_encoder = BasicEncoder(output_dim=256, norm_fn="instance")
        elif cfg.cnet == "convnext":
            self.context_encoder = convnext_large(pretrained=self.cfg.pretrain)
        elif cfg.cnet == "nat":
            self.context_encoder = nat_base(pretrained=self.cfg.pretrain)

        if cfg.pretrain_mode:
            print("[In pretrain mode, freeze context encoder]")
            for param in self.context_encoder.parameters():
                param.requires_grad = False

    def forward(self, image1, image2, mask=None, output=None, flow_init=None):
        if self.cfg.pretrain_mode:
            loss = self.pretrain_forward(image1, image2, mask=mask, output=output)
            return loss
        else:
            # Following https://github.com/princeton-vl/RAFT/
            image1 = 2 * (image1 / 255.0) - 1.0
            image2 = 2 * (image2 / 255.0) - 1.0

            data = {}

            context, _ = self.context_encoder(image1)
            context_quater = None

            (
                cost_memory,
                cost_patches,
                feat_s_quater,
                feat_t_quater,
            ) = self.memory_encoder(image1, image2, data, context)

            flow_predictions = self.memory_decoder(
                cost_memory,
                context,
                context_quater,
                feat_s_quater,
                feat_t_quater,
                data,
                flow_init=flow_init,
                cost_patches=cost_patches,
            )

            return flow_predictions

    def pretrain_forward(self, image1, image2, mask=None, output=None, flow_init=None):
        image1 = 2 * (image1 / 255.0) - 1.0
        image2 = 2 * (image2 / 255.0) - 1.0

        H_offset = self.cfg.H_offset
        W_offset = self.cfg.W_offset
        H2, W2 = self.cfg.pic_size[2:]

        image1_inner = image1[:, :, H_offset : H_offset + H2, W_offset : W_offset + W2]
        image2_inner = image2[:, :, H_offset : H_offset + H2, W_offset : W_offset + W2]

        data = {}

        context, _ = self.context_encoder(image1_inner)

        cost_memory, cost_patches = self.memory_encoder.pretrain_forward(
            image1, image2, image1_inner, image2_inner, data, context, mask=mask
        )

        loss = self.memory_decoder.pretrain_forward(
            cost_memory, context, data, flow_init=flow_init, cost_patches=cost_patches
        )

        return loss



from ..base_model.base_model import BaseModel
from .losses import MultiScale


class FlowNetBase(BaseModel):
    def __init__(
        self,
        div_flow: float = 20.0,
        input_channels: int = 6,
        batch_norm: bool = False,
        loss_start_scale: int = 4,
        loss_num_scales: int = 5,
        loss_base_weight: float = 0.32,
        loss_norm: str = "L2",
        **kwargs,
    ):
        super(FlowNetBase, self).__init__(
            loss_fn=MultiScale(
                startScale=loss_start_scale,
                numScales=loss_num_scales,
                l_weight=loss_base_weight,
                norm=loss_norm,
            ),
            output_stride=64,
            **kwargs,
        )
        self.div_flow = div_flow
        self.input_channels = input_channels
        self.batch_norm = batch_norm



import torch
import torch.nn as nn
import torch.nn.functional as F

from .update import GMAUpdateBlock
from ..encoders import twins_svt_large
from .cnn import BasicEncoder
from .corr import CorrBlock
from ..utils import coords_grid
from .gma import Attention
from .sk import SKUpdateBlock6_Deep_nopoolres_AllDecoder
from .sk2 import SKUpdateBlock6_Deep_nopoolres_AllDecoder2_Mem_predict
from .memory_util import *

try:
    from flash_attn import flash_attn_func
except:
    pass


class MemFlowNet(nn.Module):
    def __init__(self, cfg):
        super().__init__()
        self.cfg = cfg

        self.hidden_dim = 128
        self.context_dim = 128

        # feature network, context network, and update block
        if cfg.cnet == "twins":
            print("[Using twins as context encoder]")
            self.cnet = twins_svt_large(pretrained=self.cfg.pretrain)
        elif cfg.cnet == "basicencoder":
            print("[Using basicencoder as context encoder]")
            self.cnet = BasicEncoder(output_dim=256, norm_fn="batch")

        if cfg.fnet == "twins":
            print("[Using twins as feature encoder]")
            self.fnet = twins_svt_large(pretrained=self.cfg.pretrain)
        elif cfg.fnet == "basicencoder":
            print("[Using basicencoder as feature encoder]")
            self.fnet = BasicEncoder(output_dim=256, norm_fn="instance")

        if self.cfg.gma == "GMA":
            print("[Using GMA]")
            self.update_block = GMAUpdateBlock(self.cfg, hidden_dim=128)
        elif self.cfg.gma == "GMA-SK":
            print("[Using GMA-SK]")
            self.cfg.cost_heads_num = 1
            self.update_block = SKUpdateBlock6_Deep_nopoolres_AllDecoder(
                args=self.cfg, hidden_dim=128
            )
        elif self.cfg.gma == "GMA-SK2":
            print("[Using GMA-SK2]")
            self.cfg.cost_heads_num = 1
            self.update_block = SKUpdateBlock6_Deep_nopoolres_AllDecoder2_Mem_predict(
                args=self.cfg, hidden_dim=128
            )

        print("[Using corr_fn {}]".format(self.cfg.corr_fn))

        self.att = Attention(
            args=self.cfg,
            dim=self.context_dim,
            heads=1,
            max_pos_size=160,
            dim_head=self.context_dim,
        )
        self.train_avg_length = cfg.train_avg_length

        self.motion_prompt = nn.Parameter(torch.randn(1, 128, 1, 1))

    def encode_features(self, frame, flow_init=None):
        # Determine input shape
        if len(frame.shape) == 5:
            # shape is b*t*c*h*w
            need_reshape = True
            b, t = frame.shape[:2]
            # flatten so that we can feed them into a 2D CNN
            frame = frame.flatten(start_dim=0, end_dim=1)
        elif len(frame.shape) == 4:
            # shape is b*c*h*w
            need_reshape = False
        else:
            raise NotImplementedError

        fmaps = self.fnet(frame)
        if need_reshape:
            # B*T*C*H*W
            fmaps = fmaps.view(b, t, *fmaps.shape[-3:])
            frame = frame.view(b, t, *frame.shape[-3:])
            coords0, coords1 = self.initialize_flow(frame[:, 0, ...])
        else:
            coords0, coords1 = self.initialize_flow(frame)
        if flow_init is not None:
            coords1 = coords1 + flow_init

        return coords0, coords1, fmaps

    def encode_context(self, frame):
        # Determine input shape
        if len(frame.shape) == 5:
            # shape is b*t*c*h*w
            need_reshape = True
            b, t = frame.shape[:2]
            # flatten so that we can feed them into a 2D CNN
            frame = frame.flatten(start_dim=0, end_dim=1)
        elif len(frame.shape) == 4:
            # shape is b*c*h*w
            need_reshape = False
        else:
            raise NotImplementedError

        # shape is b*c*h*w
        inp = self.cnet(frame)
        inp = torch.relu(inp)
        query, key = self.att.to_qk(inp).chunk(2, dim=1)
        # query = query * self.att.scale
        if need_reshape:
            # B*C*T*H*W
            query = query.view(b, t, *query.shape[-3:]).transpose(1, 2).contiguous()
            key = key.view(b, t, *key.shape[-3:]).transpose(1, 2).contiguous()

            # B*T*C*H*W
            inp = inp.view(b, t, *inp.shape[-3:])

        return query, key, inp

    def get_motion_feature(self, flow, coords1, fmaps):
        corr_fn = CorrBlock(
            fmaps[:, 0, ...],
            fmaps[:, 1, ...],
            num_levels=self.cfg.corr_levels,
            radius=self.cfg.corr_radius,
        )
        corr = corr_fn(coords1 + flow)  # index correlation volume
        _, current_value = self.update_block.get_motion_and_value(flow, corr)
        return current_value

    def predict_flow(
        self, inp, query, ref_keys, value, forward_warp_flow=None, test_mode=False
    ):
        B, _, H, W = inp.shape
        if ref_keys is not None and value is not None:
            query = query.flatten(start_dim=2).permute(0, 2, 1).unsqueeze(2)
            ref_keys = ref_keys.flatten(start_dim=2).permute(0, 2, 1).unsqueeze(2)
            # get global motion
            # B, L, N, C
            value = value.flatten(start_dim=2).permute(0, 2, 1).unsqueeze(2)
            scale = self.att.scale * math.log(ref_keys.shape[1], self.train_avg_length)
            hidden_states = flash_attn_func(
                query, ref_keys, value, dropout_p=0.0, softmax_scale=scale, causal=False
            )
            hidden_states = (
                hidden_states.squeeze(2).permute(0, 2, 1).reshape(B, -1, H, W)
            )

            motion_features_global = (
                self.motion_prompt.repeat(B, 1, H, W)
                + self.update_block.aggregator.gamma * hidden_states
            )
        else:
            motion_features_global = self.motion_prompt.repeat(B, 1, H, W)
        if "concat_flow" in self.cfg and self.cfg.concat_flow:
            motion_features_global = torch.cat(
                [motion_features_global, forward_warp_flow], dim=1
            )
        _, up_mask, delta_flow = self.update_block(inp, motion_features_global)
        if "concat_flow" in self.cfg and self.cfg.concat_flow:
            delta_flow = delta_flow + forward_warp_flow
        # upsample predictions
        flow_up = self.upsample_flow(delta_flow, up_mask)

        return flow_up

    def initialize_flow(self, img):
        """Flow is represented as difference between two coordinate grids flow = coords1 - coords0"""
        N, C, H, W = img.shape
        coords0 = coords_grid(N, H // 8, W // 8, device=img.device, dtype=img.dtype)
        coords1 = coords_grid(N, H // 8, W // 8, device=img.device, dtype=img.dtype)

        # optical flow computed as difference: flow = coords1 - coords0
        return coords0, coords1

    def upsample_flow(self, flow, mask):
        """Upsample flow field [H/8, W/8, 2] -> [H, W, 2] using convex combination"""
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 8, 8, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(8 * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, 8 * H, 8 * W)



from typing import Sequence

import torch
import torch.nn as nn
from torch.autograd import Variable
import torch.nn.functional as F
import numpy as np

try:
    from spatial_correlation_sampler import SpatialCorrelationSampler
except ModuleNotFoundError:
    from ptlflow.utils.correlation import (
        IterSpatialCorrelationSampler as SpatialCorrelationSampler,
    )

from ptlflow.utils.registry import register_model, trainable
from .loss_functions import MultiScale_UP
from ..base_model.base_model import BaseModel


class DICL_MODULE(nn.Module):
    # Matching net with 2D conv as mentioned in the paper
    def __init__(self):
        super(DICL_MODULE, self).__init__()

        self.match = nn.Sequential(
            BasicConv(64, 96, kernel_size=3, padding=1, dilation=1),
            BasicConv(96, 128, kernel_size=3, stride=2, padding=1),  # down by 1/2
            BasicConv(128, 128, kernel_size=3, padding=1, dilation=1),
            BasicConv(128, 64, kernel_size=3, padding=1, dilation=1),
            BasicConv(
                64, 32, kernel_size=4, padding=1, stride=2, deconv=True
            ),  # up by 1/2
            nn.Conv2d(32, 1, kernel_size=3, stride=1, padding=1, bias=True),
        )

    def forward(self, x):
        x = self.match(x)
        return x


class FlowEntropy(nn.Module):
    # Compute entro from matching cost
    def __init__(self):
        super(FlowEntropy, self).__init__()

    def forward(self, x):
        x = torch.squeeze(x, 1)
        B, U, V, H, W = x.shape
        x = x.view(B, -1, H, W)
        x = F.softmax(x, dim=1).view(B, U, V, H, W)
        global_entropy = (
            (-x * torch.clamp(x, 1e-9, 1 - 1e-9).log()).sum(1).sum(1)[:, np.newaxis]
        )
        global_entropy /= np.log(x.shape[1] * x.shape[2])
        return global_entropy


class FlowRegression(nn.Module):
    # 2D soft argmin/argmax
    def __init__(self, maxU, maxV, flow_reg_by_max):
        super(FlowRegression, self).__init__()
        self.maxU = maxU
        self.maxV = maxV
        self.flow_reg_by_max = flow_reg_by_max

    def forward(self, x):
        assert x.is_contiguous() == True
        sizeU = 2 * self.maxU + 1
        sizeV = 2 * self.maxV + 1
        x = x.squeeze(1)
        B, _, _, H, W = x.shape

        # displacement along u
        dispU = torch.reshape(
            torch.arange(-self.maxU, self.maxU + 1, dtype=x.dtype),
            [1, sizeU, 1, 1, 1],
        ).to(x.device)
        dispU = dispU.expand(B, -1, sizeV, H, W).contiguous()
        dispU = dispU.view(B, sizeU * sizeV, H, W)

        # displacement along v
        dispV = torch.reshape(
            torch.arange(-self.maxV, self.maxV + 1, dtype=x.dtype),
            [1, 1, sizeV, 1, 1],
        ).to(x.device)
        dispV = dispV.expand(B, sizeU, -1, H, W).contiguous()
        dispV = dispV.view(B, sizeU * sizeV, H, W)

        x = x.view(B, sizeU * sizeV, H, W)

        if self.flow_reg_by_max:
            x = F.softmax(x, dim=1)
        else:
            x = F.softmin(x, dim=1)

        flowU = (x * dispU).sum(dim=1)
        flowV = (x * dispV).sum(dim=1)
        flow = torch.cat((flowU.unsqueeze(1), flowV.unsqueeze(1)), dim=1)
        return flow


class DAP(nn.Module):
    def __init__(self, md=3, dap_by_temperature=False):
        # Displacement-aware projection layer
        # implemented as a 1x1 2D conv
        super(DAP, self).__init__()
        self.dap_by_temperature = dap_by_temperature
        dimC = (2 * md + 1) ** 2
        self.dap_layer = BasicConv(
            dimC, dimC, kernel_size=1, padding=0, stride=1, bn=False, relu=False
        )
        if self.dap_by_temperature:
            self.dap_layer = BasicConv(
                dimC, 1, kernel_size=1, padding=0, stride=1, bn=False, relu=False
            )

    def forward(self, x):
        x = x.squeeze(1)
        bs, du, dv, h, w = x.shape
        x = x.view(bs, du * dv, h, w)

        if self.dap_by_temperature:
            temp = self.dap_layer(x) + 1e-6
            x = x * temp
        else:
            x = self.dap_layer(x)

        return x.view(bs, du, dv, h, w).unsqueeze(1)


class DICL(BaseModel):
    pretrained_checkpoints = {
        "chairs": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/dicl-chairs-fdc24e2f.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/dicl-kitti-4813ccab.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/dicl-sintel-fa9fc259.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/dicl-things-20bc573d.ckpt",
    }

    def __init__(
        self,
        ctf_context: bool = True,
        ctf_context_only_flow2: bool = False,
        cuda_cost: bool = False,
        dap_by_temperature: bool = False,
        dap_layer: bool = True,
        dap_init_by_id: bool = True,
        flow_reg_by_max: bool = True,
        remove_warp_hole: bool = True,
        scale_context: Sequence[float] = (0.03125, 0.0625, 0.125, 0.25, 0.5),
        search_range: Sequence[int] = (3, 3, 3, 3, 3),
        sup_raw_flow: bool = False,
        loss_removezero: bool = False,
        loss_type: str = "L1",
        loss_use_valid_range: bool = False,
        loss_weight: Sequence[float] = (1.0, 0.75, 0.50, 0.25, 0.25),
        loss_valid_range: Sequence[tuple[int, int]] = (
            (372, 372),
            (360, 360),
            (336, 336),
            (288, 288),
            (192, 192),
        ),
        pad_by_cons: bool = False,
        **kwargs,
    ):
        super(DICL, self).__init__(
            loss_fn=MultiScale_UP(
                loss_type=loss_type,
                weight=loss_weight,
                valid_range=loss_valid_range,
                removezero=loss_removezero,
                use_valid_range=loss_use_valid_range,
            ),
            output_stride=128,
            **kwargs,
        )

        self.ctf_context = ctf_context
        self.ctf_context_only_flow2 = ctf_context_only_flow2
        self.cuda_cost = cuda_cost
        self.dap_by_temperature = dap_by_temperature
        self.dap_layer = dap_layer
        self.dap_init_by_id = dap_init_by_id
        self.flow_reg_by_max = flow_reg_by_max
        self.remove_warp_hole = remove_warp_hole
        self.scale_context = scale_context
        self.search_range = search_range
        self.sup_raw_flow = sup_raw_flow
        self.pad_by_cons = pad_by_cons

        self.feature = FeatureGA()

        if self.dap_layer:
            self.dap_layer6 = DAP(
                md=self.search_range[4],
                dap_by_temperature=self.dap_by_temperature,
            )
            self.dap_layer5 = DAP(
                md=self.search_range[3],
                dap_by_temperature=self.dap_by_temperature,
            )
            self.dap_layer4 = DAP(
                md=self.search_range[2],
                dap_by_temperature=self.dap_by_temperature,
            )
            self.dap_layer3 = DAP(
                md=self.search_range[1],
                dap_by_temperature=self.dap_by_temperature,
            )
            self.dap_layer2 = DAP(
                md=self.search_range[0],
                dap_by_temperature=self.dap_by_temperature,
            )
        else:
            self.dap_layer6 = self.dap_layer5 = self.dap_layer4 = self.dap_layer3 = (
                self.dap_layer2
            ) = None

        self.entropy = FlowEntropy()

        # matching net, with the same arch at each pyramid level
        # level 6->2:  1/64,1/32,1/16,1/8,1/4
        self.matching6 = DICL_MODULE()
        self.matching5 = DICL_MODULE()
        self.matching4 = DICL_MODULE()
        self.matching3 = DICL_MODULE()
        self.matching2 = DICL_MODULE()

        # search range, e.g., [-3,3]
        # the search range for FlowRegression should be aligned with that when computing matching cost
        self.flow6 = FlowRegression(
            self.search_range[4],
            self.search_range[4],
            self.flow_reg_by_max,
        )
        self.flow5 = FlowRegression(
            self.search_range[3],
            self.search_range[3],
            self.flow_reg_by_max,
        )
        self.flow4 = FlowRegression(
            self.search_range[2],
            self.search_range[2],
            self.flow_reg_by_max,
        )
        self.flow3 = FlowRegression(
            self.search_range[1],
            self.search_range[1],
            self.flow_reg_by_max,
        )
        self.flow2 = FlowRegression(
            self.search_range[0],
            self.search_range[0],
            self.flow_reg_by_max,
        )

        if ctf_context:
            # If you would use context network as introduced in PWCNet
            self.context_net2 = nn.Sequential(
                BasicConv(38, 64, kernel_size=3, padding=1, dilation=1),
                BasicConv(64, 128, kernel_size=3, padding=2, dilation=2),
                BasicConv(128, 128, kernel_size=3, padding=4, dilation=4),
                BasicConv(128, 96, kernel_size=3, padding=8, dilation=8),
                BasicConv(96, 64, kernel_size=3, padding=16, dilation=16),
                BasicConv(64, 32, kernel_size=3, padding=1, dilation=1),
                nn.Conv2d(32, 2, kernel_size=3, stride=1, padding=1, bias=True),
            )
            self.context_net3 = nn.Sequential(
                BasicConv(38, 64, kernel_size=3, padding=1, dilation=1),
                BasicConv(64, 128, kernel_size=3, padding=2, dilation=2),
                BasicConv(128, 128, kernel_size=3, padding=4, dilation=4),
                BasicConv(128, 96, kernel_size=3, padding=8, dilation=8),
                BasicConv(96, 64, kernel_size=3, padding=16, dilation=16),
                BasicConv(64, 32, kernel_size=3, padding=1, dilation=1),
                nn.Conv2d(32, 2, kernel_size=3, stride=1, padding=1, bias=True),
            )
            self.context_net4 = nn.Sequential(
                BasicConv(38, 64, kernel_size=3, padding=1, dilation=1),
                BasicConv(64, 128, kernel_size=3, padding=2, dilation=2),
                BasicConv(128, 128, kernel_size=3, padding=4, dilation=4),
                BasicConv(128, 64, kernel_size=3, padding=8, dilation=8),
                BasicConv(64, 32, kernel_size=3, padding=1, dilation=1),
                nn.Conv2d(32, 2, kernel_size=3, stride=1, padding=1, bias=True),
            )
            self.context_net5 = nn.Sequential(
                BasicConv(38, 64, kernel_size=3, padding=1, dilation=1),
                BasicConv(64, 128, kernel_size=3, padding=2, dilation=2),
                BasicConv(128, 64, kernel_size=3, padding=4, dilation=4),
                BasicConv(64, 32, kernel_size=3, padding=1, dilation=1),
                nn.Conv2d(32, 2, kernel_size=3, stride=1, padding=1, bias=True),
            )
            # We remove the last several layers on pyramid level 6,5,4 considering their inputs' resolution
            # though this does not have an obvious effect
            self.context_net6 = nn.Sequential(
                BasicConv(38, 64, kernel_size=3, padding=1, dilation=1),
                BasicConv(64, 64, kernel_size=3, padding=2, dilation=2),
                BasicConv(64, 32, kernel_size=3, padding=1, dilation=1),
                nn.Conv2d(32, 2, kernel_size=3, stride=1, padding=1, bias=True),
            )

        for m in self.modules():
            if isinstance(m, (nn.Conv2d)) or isinstance(m, nn.ConvTranspose2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, (nn.BatchNorm2d)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        if self.dap_init_by_id:
            # init the dap layer kernel by identity matrix
            if self.dap_layer6 is not None:
                nn.init.eye_(
                    self.dap_layer6.dap_layer.conv.weight.squeeze(-1).squeeze(-1)
                )
            if self.dap_layer5 is not None:
                nn.init.eye_(
                    self.dap_layer5.dap_layer.conv.weight.squeeze(-1).squeeze(-1)
                )
            if self.dap_layer4 is not None:
                nn.init.eye_(
                    self.dap_layer4.dap_layer.conv.weight.squeeze(-1).squeeze(-1)
                )
            if self.dap_layer3 is not None:
                nn.init.eye_(
                    self.dap_layer3.dap_layer.conv.weight.squeeze(-1).squeeze(-1)
                )
            if self.dap_layer2 is not None:
                nn.init.eye_(
                    self.dap_layer2.dap_layer.conv.weight.squeeze(-1).squeeze(-1)
                )

    def warp(self, x, flo):
        """
        warp an image/tensor (im2) back to im1, according to the optical flow
        x: [B, C, H, W] (im2)
        flo: [B, 2, H, W] flow
        """
        B, C, H, W = x.size()
        # mesh grid
        xx = torch.arange(0, W, dtype=x.dtype, device=x.device).view(1, -1).repeat(H, 1)
        yy = torch.arange(0, H, dtype=x.dtype, device=x.device).view(-1, 1).repeat(1, W)
        xx = xx.view(1, 1, H, W).repeat(B, 1, 1, 1)
        yy = yy.view(1, 1, H, W).repeat(B, 1, 1, 1)
        grid = torch.cat((xx, yy), 1)

        if x.is_cuda:
            grid = grid.cuda()
        vgrid = Variable(grid) + flo

        # scale grid to [-1,1]
        vgrid[:, 0, :, :] = 2.0 * vgrid[:, 0, :, :] / max(W - 1, 1) - 1.0
        vgrid[:, 1, :, :] = 2.0 * vgrid[:, 1, :, :] / max(H - 1, 1) - 1.0

        vgrid = vgrid.permute(0, 2, 3, 1)
        output = nn.functional.grid_sample(x, vgrid, align_corners=True)
        mask = torch.autograd.Variable(
            torch.ones(x.size(), dtype=x.dtype, device=x.device)
        )
        mask = nn.functional.grid_sample(mask, vgrid, align_corners=True)

        mask[mask < 0.9999] = 0
        mask[mask > 0] = 1

        return output * mask, mask

    def forward(self, inputs):
        # left and right images normalized
        if self.pad_by_cons:
            pad_mode = "constant"
            pad_value = 1
        else:
            pad_mode = "replicate"
            pad_value = None

        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=True,
            resize_mode="pad",
            pad_mode=pad_mode,
            pad_value=pad_value,
            pad_two_side=False,
        )

        x = images[:, 0]
        y = images[:, 1]

        # feature extraction
        _, x2, x3, x4, x5, x6 = self.feature(x)
        _, y2, y3, y4, y5, y6 = self.feature(y)

        # compute flow for level 6
        cost6 = self.compute_cost(
            x6, y6, self.matching6, self.search_range[4], self.search_range[4]
        )
        g6 = F.interpolate(
            x,
            [x.shape[-2] // 64, x.shape[-1] // 64],
            mode="bilinear",
            align_corners=True,
        )
        if self.dap_layer6 is not None:
            cost6 = self.dap_layer6(cost6)
        flow6 = self.flow6(cost6)
        if self.ctf_context:
            if self.sup_raw_flow:
                raw_flow6 = flow6
            entro6 = self.entropy(cost6)
            # input to context network
            # including pred flow, entro, left img feat, left img
            feat6 = torch.cat((flow6.detach(), entro6.detach(), x6, g6), dim=1)
            flow6 = flow6 + self.context_net6(feat6) * self.scale_context[0]
        up_flow6 = 2.0 * F.interpolate(
            flow6, [x5.shape[2], x5.shape[3]], mode="bilinear", align_corners=True
        )
        up_flow6 = up_flow6.detach()

        # compute flow for level 5
        warp5, _ = self.warp(y5, up_flow6)
        cost5 = self.compute_cost(
            x5,
            warp5,
            self.matching5,
            self.search_range[3],
            self.search_range[3],
        )
        g5 = F.interpolate(
            x,
            [x.shape[-2] // 32, x.shape[-1] // 32],
            mode="bilinear",
            align_corners=True,
        )
        if self.dap_layer5 is not None:
            cost5 = self.dap_layer5(cost5)
        flow5 = self.flow5(cost5) + up_flow6
        if self.ctf_context:
            if self.sup_raw_flow:
                raw_flow5 = flow5
            entro5 = self.entropy(cost5)
            feat5 = torch.cat((flow5.detach(), entro5.detach(), x5, g5), dim=1)
            flow5 = flow5 + self.context_net5(feat5) * self.scale_context[1]
        up_flow5 = 2.0 * F.interpolate(
            flow5, [x4.shape[2], x4.shape[3]], mode="bilinear", align_corners=True
        )
        up_flow5 = up_flow5.detach()

        # compute flow for level 4
        warp4, _ = self.warp(y4, up_flow5)
        cost4 = self.compute_cost(
            x4,
            warp4,
            self.matching4,
            self.search_range[2],
            self.search_range[2],
        )
        g4 = F.interpolate(
            x,
            [x.shape[-2] // 16, x.shape[-1] // 16],
            mode="bilinear",
            align_corners=True,
        )
        if self.dap_layer4 is not None:
            cost4 = self.dap_layer4(cost4)
        flow4 = self.flow4(cost4) + up_flow5
        if self.ctf_context:
            if self.sup_raw_flow:
                raw_flow4 = flow4
            entro4 = self.entropy(cost4)
            feat4 = torch.cat((flow4.detach(), entro4.detach(), x4, g4), dim=1)
            flow4 = flow4 + self.context_net4(feat4) * self.scale_context[2]
        up_flow4 = 2.0 * F.interpolate(
            flow4, [x3.shape[2], x3.shape[3]], mode="bilinear", align_corners=True
        )
        up_flow4 = up_flow4.detach()

        # compute flow for level 3
        warp3, _ = self.warp(y3, up_flow4)
        cost3 = self.compute_cost(
            x3,
            warp3,
            self.matching3,
            self.search_range[1],
            self.search_range[1],
        )
        g3 = F.interpolate(
            x, [x.shape[-2] // 8, x.shape[-1] // 8], mode="bilinear", align_corners=True
        )
        if self.dap_layer3 is not None:
            cost3 = self.dap_layer3(cost3)
        flow3 = self.flow3(cost3) + up_flow4
        if self.ctf_context:
            if self.sup_raw_flow:
                raw_flow3 = flow3
            entro3 = self.entropy(cost3)
            feat3 = torch.cat((flow3.detach(), entro3.detach(), x3, g3), dim=1)
            flow3 = flow3 + self.context_net3(feat3) * self.scale_context[3]
        up_flow3 = 2.0 * F.interpolate(
            flow3, [x2.shape[2], x2.shape[3]], mode="bilinear", align_corners=True
        )
        up_flow3 = up_flow3.detach()

        # compute flow for level 2
        warp2, _ = self.warp(y2, up_flow3)
        cost2 = self.compute_cost(
            x2,
            warp2,
            self.matching2,
            self.search_range[0],
            self.search_range[0],
        )
        g2 = F.interpolate(
            x, [x.shape[-2] // 4, x.shape[-1] // 4], mode="bilinear", align_corners=True
        )
        if self.dap_layer2 is not None:
            cost2 = self.dap_layer2(cost2)
        flow2 = self.flow2(cost2) + up_flow3

        if self.ctf_context or self.ctf_context_only_flow2:
            if self.sup_raw_flow:
                raw_flow2 = flow2
            entro2 = self.entropy(cost2)
            feat2 = torch.cat((flow2.detach(), entro2.detach(), x2, g2), dim=1)
            flow2 = flow2 + self.context_net2(feat2) * self.scale_context[4]
        # sds

        flow_out = F.interpolate(
            flow2, size=x.shape[-2:], mode="bilinear", align_corners=True
        )[:, None]
        flow_out[:, :, 0] = flow_out[:, :, 0] * (float(x.shape[-1]) / flow2.shape[-1])
        flow_out[:, :, 1] = flow_out[:, :, 1] * (float(x.shape[-2]) / flow2.shape[-2])
        flow_out = self.postprocess_predictions(flow_out, image_resizer, is_flow=True)

        output = {"flows": flow_out}
        if self.training:
            if self.sup_raw_flow:
                output["flow_preds"] = [
                    flow2,
                    raw_flow2,
                    flow3,
                    raw_flow3,
                    flow4,
                    raw_flow4,
                    flow5,
                    raw_flow5,
                    flow6,
                    raw_flow6,
                ]
            else:
                output["flow_preds"] = [flow2, flow3, flow4, flow5, flow6]
        return output

    def compute_cost(self, x, y, matchnet, maxU, maxV):
        sizeU = 2 * maxU + 1
        sizeV = 2 * maxV + 1
        b, c, height, width = x.shape

        with torch.cuda.device_of(x):
            # init cost as tensor matrix
            cost = (
                x.new()
                .resize_(x.size()[0], 2 * c, 2 * maxU + 1, 2 * maxV + 1, height, width)
                .zero_()
            )

        if self.cuda_cost:
            # CUDA acceleration
            corr = SpatialCorrelationSampler(
                kernel_size=1,
                patch_size=(int(1 + 2 * 3), int(1 + 2 * 3)),
                stride=1,
                padding=0,
                dilation_patch=1,
            )
            cost = corr(x, y)
        else:
            for i in range(2 * maxU + 1):
                ind = i - maxU
                for j in range(2 * maxV + 1):
                    indd = j - maxV
                    # for each displacement hypothesis, we construct a feature map as the input of matching net
                    # here we hold them together for parallel processing later
                    cost[
                        :,
                        :c,
                        i,
                        j,
                        max(0, -indd) : height - indd,
                        max(0, -ind) : width - ind,
                    ] = x[
                        :, :, max(0, -indd) : height - indd, max(0, -ind) : width - ind
                    ]
                    cost[
                        :,
                        c:,
                        i,
                        j,
                        max(0, -indd) : height - indd,
                        max(0, -ind) : width - ind,
                    ] = y[
                        :, :, max(0, +indd) : height + indd, max(0, ind) : width + ind
                    ]

        if self.remove_warp_hole:
            # mitigate the effect of holes (may be raised by occ)
            valid_mask = cost[:, c:, ...].sum(dim=1) != 0
            valid_mask = valid_mask.detach()
            cost = cost * valid_mask.unsqueeze(1)

        # (B, 2C, U, V, H, W) -> (B, U, V, 2C, H, W)
        cost = cost.permute([0, 2, 3, 1, 4, 5]).contiguous()
        # (B, U, V, 2C, H, W) -> (BxUxV, 2C, H, W)
        cost = cost.view(x.size()[0] * sizeU * sizeV, c * 2, x.size()[2], x.size()[3])
        # (BxUxV, 2C, H, W) -> (BxUxV, 1, H, W)
        cost = matchnet(cost)
        cost = cost.view(x.size()[0], sizeU, sizeV, 1, x.size()[2], x.size()[3])
        cost = cost.permute([0, 3, 1, 2, 4, 5]).contiguous()

        # (B, U, V, H, W)
        return cost


######################## Part of the Model #############################################


class ResBlock(nn.Module):
    def __init__(self, in_channel, out_channel, dilation=1, stride=1, downsample=None):
        super(ResBlock, self).__init__()

        # To keep the shape of input and output same when dilation conv, we should compute the padding:
        # Reference:
        #   https://discuss.pytorch.org/t/how-to-keep-the-shape-of-input-and-output-same-when-dilation-conv/14338
        # padding = [(o-1)*s+k+(k-1)*(d-1)-i]/2, here the i is input size, and o is output size.
        # set o = i, then padding = [i*(s-1)+k+(k-1)*(d-1)]/2 = [k+(k-1)*(d-1)]/2      , stride always equals 1
        # if dilation != 1:
        #     padding = (3+(3-1)*(dilation-1))/2
        padding = dilation

        self.conv1 = nn.Conv2d(
            in_channel,
            out_channel,
            kernel_size=3,
            stride=stride,
            padding=padding,
            dilation=dilation,
            bias=False,
        )
        self.bn1 = nn.BatchNorm2d(out_channel)
        self.relu1 = nn.LeakyReLU(negative_slope=0.1, inplace=True)

        self.conv2 = nn.Conv2d(
            out_channel,
            out_channel,
            kernel_size=3,
            stride=stride,
            padding=padding,
            dilation=dilation,
            bias=False,
        )
        self.bn2 = nn.BatchNorm2d(out_channel)
        self.relu2 = nn.LeakyReLU(negative_slope=0.1, inplace=True)

        self.downsample = downsample
        self.stride = stride
        self.in_ch = in_channel
        self.out_ch = out_channel
        self.p = padding
        self.d = dilation

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)

        out = self.relu1(out)
        out = self.conv2(out)
        out = self.bn2(out)

        out += residual

        out = self.relu2(out)

        return out


class BasicConv(nn.Module):
    def __init__(
        self, in_channels, out_channels, deconv=False, bn=True, relu=True, **kwargs
    ):
        super(BasicConv, self).__init__()
        self.relu = relu
        self.use_bn = bn
        if self.use_bn:
            self.bn = nn.BatchNorm2d(out_channels)
        if deconv:
            self.conv = nn.ConvTranspose2d(
                in_channels, out_channels, bias=False, **kwargs
            )
        else:
            self.conv = nn.Conv2d(in_channels, out_channels, bias=False, **kwargs)

    def forward(self, x):
        x = self.conv(x)
        if self.use_bn:
            x = self.bn(x)
        if self.relu:
            x = F.relu(x, inplace=True)
        return x


class Conv2x(nn.Module):
    def __init__(
        self, in_channels, out_channels, deconv=False, concat=True, bn=True, relu=True
    ):
        super(Conv2x, self).__init__()
        self.concat = concat
        self.deconv = deconv
        if deconv:
            kernel = 4
        else:
            kernel = 3

        self.conv1 = BasicConv(
            in_channels,
            out_channels,
            deconv,
            bn=False,
            relu=True,
            kernel_size=kernel,
            stride=2,
            padding=1,
        )

        if self.concat:
            self.conv2 = BasicConv(
                out_channels * 2,
                out_channels,
                False,
                bn,
                relu,
                kernel_size=3,
                stride=1,
                padding=1,
            )
        else:
            self.conv2 = BasicConv(
                out_channels,
                out_channels,
                False,
                bn,
                relu,
                kernel_size=3,
                stride=1,
                padding=1,
            )

    def forward(self, x, rem):
        x = self.conv1(x)
        assert x.size() == rem.size()

        if self.concat:
            x = torch.cat((x, rem), 1)
        else:
            x = x + rem
        x = self.conv2(x)
        return x


class FeatureGA(nn.Module):
    def __init__(self):
        super(FeatureGA, self).__init__()
        # feature backbone
        # adopted from GANet (https://github.com/feihuzhang/GANet)
        self.conv_start = nn.Sequential(
            BasicConv(3, 32, kernel_size=3, padding=1),
            BasicConv(32, 32, kernel_size=3, stride=2, padding=1),
            BasicConv(32, 32, kernel_size=3, padding=1),
        )
        self.conv1a = BasicConv(32, 48, kernel_size=3, stride=2, padding=1)
        self.conv2a = BasicConv(48, 64, kernel_size=3, stride=2, padding=1)
        self.conv3a = BasicConv(64, 96, kernel_size=3, stride=2, padding=1)
        self.conv4a = BasicConv(96, 128, kernel_size=3, stride=2, padding=1)
        self.conv5a = BasicConv(128, 160, kernel_size=3, stride=2, padding=1)
        self.conv6a = BasicConv(160, 192, kernel_size=3, stride=2, padding=1)

        self.deconv6a = Conv2x(192, 160, deconv=True)
        self.deconv5a = Conv2x(160, 128, deconv=True)
        self.deconv4a = Conv2x(128, 96, deconv=True)
        self.deconv3a = Conv2x(96, 64, deconv=True)
        self.deconv2a = Conv2x(64, 48, deconv=True)
        self.deconv1a = Conv2x(48, 32, deconv=True)

        self.conv1b = Conv2x(32, 48)
        self.conv2b = Conv2x(48, 64)
        self.conv3b = Conv2x(64, 96)
        self.conv4b = Conv2x(96, 128)
        self.conv5b = Conv2x(128, 160)
        self.conv6b = Conv2x(160, 192)

        self.deconv6b = Conv2x(192, 160, deconv=True)
        self.outconv_6 = BasicConv(160, 32, kernel_size=3, padding=1)

        self.deconv5b = Conv2x(160, 128, deconv=True)
        self.outconv_5 = BasicConv(128, 32, kernel_size=3, padding=1)

        self.deconv4b = Conv2x(128, 96, deconv=True)
        self.outconv_4 = BasicConv(96, 32, kernel_size=3, padding=1)

        self.deconv3b = Conv2x(96, 64, deconv=True)
        self.outconv_3 = BasicConv(64, 32, kernel_size=3, padding=1)

        self.deconv2b = Conv2x(64, 48, deconv=True)
        self.outconv_2 = BasicConv(48, 32, kernel_size=3, padding=1)

    def forward(self, x):
        x = self.conv_start(x)
        rem0 = x
        x = self.conv1a(x)
        rem1 = x
        x = self.conv2a(x)
        rem2 = x
        x = self.conv3a(x)
        rem3 = x
        x = self.conv4a(x)
        rem4 = x
        x = self.conv5a(x)
        rem5 = x
        x = self.conv6a(x)
        rem6 = x

        x = self.deconv6a(x, rem5)
        rem5 = x
        x = self.deconv5a(x, rem4)
        rem4 = x
        x = self.deconv4a(x, rem3)
        rem3 = x
        x = self.deconv3a(x, rem2)
        rem2 = x
        x = self.deconv2a(x, rem1)
        rem1 = x
        x = self.deconv1a(x, rem0)
        rem0 = x

        x = self.conv1b(x, rem1)
        rem1 = x
        x = self.conv2b(x, rem2)
        rem2 = x
        x = self.conv3b(x, rem3)
        rem3 = x
        x = self.conv4b(x, rem4)
        rem4 = x
        x = self.conv5b(x, rem5)
        rem5 = x
        x = self.conv6b(x, rem6)

        x = self.deconv6b(x, rem5)
        x6 = self.outconv_6(x)

        x = self.deconv5b(x, rem4)
        x5 = self.outconv_5(x)

        x = self.deconv4b(x, rem3)
        x4 = self.outconv_4(x)

        x = self.deconv3b(x, rem2)
        x3 = self.outconv_3(x)

        x = self.deconv2b(x, rem1)
        x2 = self.outconv_2(x)

        return None, x2, x3, x4, x5, x6


@register_model
@trainable
class dicl(DICL):
    pass



try:
    import faiss

    res = faiss.StandardGpuResources()
    res.setDefaultNullStreamAllDevices()
except ImportError:
    faiss = None
import torch


def swig_ptr_from_Tensor(x):
    """gets a Faiss SWIG pointer from a pytorch tensor (on CPU or GPU)"""
    assert x.is_contiguous()

    if x.dtype == torch.float32:
        return faiss.cast_integer_to_float_ptr(
            x.untyped_storage().data_ptr() + x.storage_offset() * 4
        )

    if x.dtype == torch.int64:
        return faiss.cast_integer_to_int_ptr(
            x.untyped_storage().data_ptr() + x.storage_offset() * 8
        )

    raise Exception("tensor type not supported: {}".format(x.dtype))


def search_raw_array_pytorch(
    res, xb, xq, k, D=None, I=None, metric=faiss.METRIC_L2 if faiss is not None else 0
):
    """search xq in xb, without building an index"""
    assert xb.device == xq.device

    nq, d = xq.size()
    if xq.is_contiguous():
        xq_row_major = True
    elif xq.t().is_contiguous():
        xq = xq.t()  # I initially wrote xq:t(), Lua is still haunting me :-)
        xq_row_major = False
    else:
        raise TypeError("matrix should be row or column-major")

    xq_ptr = swig_ptr_from_Tensor(xq)

    nb, d2 = xb.size()
    assert d2 == d
    if xb.is_contiguous():
        xb_row_major = True
    elif xb.t().is_contiguous():
        xb = xb.t()
        xb_row_major = False
    else:
        raise TypeError("matrix should be row or column-major")
    xb_ptr = swig_ptr_from_Tensor(xb)

    if D is None:
        D = torch.empty(nq, k, device=xb.device, dtype=torch.float32)
    else:
        assert D.shape == (nq, k)
        assert D.device == xb.device

    if I is None:
        I = torch.empty(nq, k, device=xb.device, dtype=torch.int64)
    else:
        assert I.shape == (nq, k)
        assert I.device == xb.device

    D_ptr = swig_ptr_from_Tensor(D)
    I_ptr = swig_ptr_from_Tensor(I)

    args = faiss.GpuDistanceParams()
    args.metric = metric
    args.k = k
    args.dims = d
    args.vectors = xb_ptr
    args.vectorsRowMajor = xb_row_major
    args.numVectors = nb
    args.queries = xq_ptr
    args.queriesRowMajor = xq_row_major
    args.numQueries = nq
    args.outDistances = D_ptr
    args.outIndices = I_ptr
    faiss.bfKnn(res, args)

    return D, I


def knn_faiss_raw(fmap1, fmap2, k):
    b, ch, _ = fmap1.shape

    if b == 1:
        fmap1 = fmap1.view(ch, -1).t().contiguous()
        fmap2 = fmap2.view(ch, -1).t().contiguous()

        dist, indx = search_raw_array_pytorch(
            res, fmap2, fmap1, k, metric=faiss.METRIC_INNER_PRODUCT
        )

        dist = dist.t().unsqueeze(0).contiguous()
        indx = indx.t().unsqueeze(0).contiguous()
    else:
        fmap1 = fmap1.view(b, ch, -1).permute(0, 2, 1).contiguous()
        fmap2 = fmap2.view(b, ch, -1).permute(0, 2, 1).contiguous()
        dist = []
        indx = []
        for i in range(b):
            dist_i, indx_i = search_raw_array_pytorch(
                res, fmap2[i], fmap1[i], k, metric=faiss.METRIC_INNER_PRODUCT
            )
            dist_i = dist_i.t().unsqueeze(0).contiguous()
            indx_i = indx_i.t().unsqueeze(0).contiguous()
            dist.append(dist_i)
            indx.append(indx_i)
        dist = torch.cat(dist, dim=0)
        indx = torch.cat(indx, dim=0)
    return dist, indx
